// VinciaRoot.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for the Vincia ROOT utilities

#ifndef VINCIAROOT_H
#define VINCIAROOT_H

// Include specific ROOT headers 

// Main ROOT header
#include "TROOT.h"
// 1D and 2D histograms
#include "TH1.h"
#include "TH2.h"
#include "TGraphAsymmErrors.h"
#include "TPaveStats.h"
#include "TLatex.h"
#include "TLegend.h"
#include "TLegendEntry.h"
#include "TLine.h"
// Widget windows
#include "TCanvas.h"
#include "TApplication.h"
// To read/write files
#include "TDirectory.h"
#include "TFile.h"

// Put all the rest in Vincia namespace

namespace Vincia {
  
  //***************************************************************
  // class UncertaintyBand: 
  // Extremely simple auxiliary class for storing names and indices 
  // of uncertainty bands
  //***************************************************************

  class UncertaintyBand {
    
  public: 
    
    // Constructor
    UncertaintyBand(string nameIn=" ") {
      name = nameIn;
      indices.resize(0);
    };
    
    // Name
    string name;
    // Indices
    vector<int> indices;
    
  };
  
  //***************************************************************
  // class VinciaRoot: Main VinciaRoot Class
  // Note: is completely independent of VINCIA and can be used 
  // with any executable using ROOT histograms (specifically TH1F)
  //***************************************************************

  class VinciaRoot {

  public:

    // Constructor
    VinciaRoot() {

      // Switch off display by default
      useDisplay = false;
      appIsInit  = false;

      // Switch on storage of stat uncertainties 
      TH1::SetDefaultSumw2();

      // Default verbose level
      verbose = 1;

      //--------------------------------------------------------------
      // Data Colors are hardcoded for the time being 
      dataColor     =  1; 
      // Data inner color (stat)
      dataStatColor = 90;
      // Data outer color
      dataSysColor  = 91;
      
      // Default user colors
      userColor1 = 61;
      userColor2 = 64;
      userColor3 = 66;
      userColor4 = 67;
      refColor   =  2;

      // Default refs
      mcRefs  = " ";
      mcLabel = "MC";

      // Default stat options
      optStat = 1110;

    }
    
    //------------------------------------------------------------------------
    // createDisplays
    // Initialize Displays

    bool createDisplay(string nameIn="VINCIA", int xSizIn=1024, int ySizIn=800, int xOffsetIn=-1,int yOffsetIn=-1) {
      //      cout<<"createDisplay..."<<endl;
      displayName.push_back(nameIn);
      xSiz.push_back(xSizIn);
      ySiz.push_back(ySizIn);      
      if (xOffsetIn == -1) xOffsetIn = 16*xOffset.size();
      if (yOffsetIn == -1) yOffsetIn = 16*yOffset.size();
      xOffset.push_back(xOffsetIn);
      yOffset.push_back(yOffsetIn);      
      nColsSav.push_back(0);
      nRowsSav.push_back(0);
      if (! appIsInit) {
	cout<<"(VinciaRoot:) initializing app"<<endl;
	useDisplay=true;
	gROOT->Reset();
	gROOT->SetStyle("Plain");	
	TApplication *vcApp = new TApplication("vcApp",0,0);
	vcApp->InitializeGraphics();
	appIsInit = true;
      }      
      //cout<<"  adding display"<<endl;
      displays.push_back(new TCanvas(nameIn.c_str(),nameIn.c_str(),
				     xOffsetIn,       yOffsetIn,
				     xSizIn,ySizIn));
      displays[displays.size()-1]->Modified();
      displays[displays.size()-1]->Update();      
      
      //cout<<"  returning";
      return true;
    }
    
    //------------------------------------------------------------------------
    // addDataFile
    // Read file containing data and store points
    // Needs to be modified for each new data file, if using different
    // formats.

    bool addDataFile(string dataFile, bool normalize=true) {

      // Define read-in stream
      ifstream in;

      // Index
      int iHist = vTheory.size()-1;

      // Return false if fails
      in.open(dataFile.c_str());
      if (! in) return false;

      // Arrays for data storage
      Float_t xc,xlo,xhi,yc,sig1(0.),sig2(0.),sig3(0.),sig4(0.),sig5(0.),sig6(0.);
      Int_t nlines = 0;
      Float_t x[200],exL[200],exR[200],xT[200],exLT[200],exRT[200];
      Float_t y[200],eyStatL[200],eyStatH[200],eyTotL[200],eyTotH[200], zero[200];
      Float_t yT[200],eyStatLT[200],eyStatHT[200],eyTotLT[200],eyTotHT[200];
      Float_t sumsig=0.0;
      
      // Set data format depending on which file we're reading
      int    dataFormat = 11;
      // Don't apply the TH norm factor to the data
      // I.e., the TH norm factor gives a way of normalizing TH differently 
      double normFactor = 1.0; //vTheory[iHist]->GetNormFactor();     
      // if (normFactor <= 0.0) normFactor = 1.0; 
      bool normalizeData = normalize;
      // Format  0: x, y, yErr
      //           with bin widths fixed to xc +/- 0.5
      // Format  1: x, y, sigStat, sigSys
      //           with bin widths fixed to xc +/- 1 (or max to next)
      // Format  2: xlo, xhi, y, sigStatP, sigStatM, sigSysP, sigSysM
      //           with bin widths fixed to xc +/- 1 
      // Format  3: xlo, xhi, y, sigStatP, sigStatM, sigSysP, sigSysM
      //           with dynamic bin widths (bins assumed symmetric)
      // Format  4: xlo, xhi, y, sigStatPM, sigSysP, sigSysM
      //           with dynamic bin widths (bins assumed symmetric)
      // Format  5: xlo, xhi, y, sigStatPM, sigSys1PM, sigSys2PM
      // Format  6: xlo, xhi, y, sigTotPM, sigStatPM, sigSysPM
      // Format  7: xlo, xhi, y, sigStat, sigSys
      // Format  8: xlo, xhi, y, sig1P, sig1M, sig2P, sig2M, sig3P, sig3M
      // Format -7: exp(xlo), exp(xhi), y, sigStat, sigSys
      // Format  9: x, xlo, xhi, y, sigStatP, sigStatM, sig2P, sig2M
      // SLD B fragmentation measurements
      // Format 10: x, xlo, xhi, y, sigStatP, sigStatM, sig2P, sig2M, sig3P, sig3M
      // mcplots formats
      // Format 11: xlo, xx, xhi, y, sigP, sigM
      // Format 12: xlo, xx, xhi, y, sigP, sigM, sigP2, sigM2
      // Format 13: xlo, xx, xhi, y, sigP, sigM, sigP2, sigM2, sigP3, sigP3

      // DELPHI DATA
      if (dataFile.find("Delphi") != string::npos) {
	dataFormat = 3;
	if (dataFile.find("a34") != string::npos ||
	    dataFile.find("chiBZ") != string::npos ||
	    dataFile.find("thetaNR") != string::npos ||
	    dataFile.find("phiKSW") != string::npos )
	  dataFormat = 7;
	if (dataFile.find("R3bl") != string::npos) {
	  dataFormat = 1;
	  normalizeData = false;
	}
      } 
      // ALEPH DATA
      else if (dataFile.find("Aleph") != string::npos) {
	dataFormat = 3;
	if (dataFile.find("xK") != string::npos) {
	  dataFormat = -7;
	  normalizeData = false;
	} 
	else if (dataFile.find("-r") != string::npos && dataFile.find("j.") !=
		 string::npos) {
	  dataFormat = 8;
	}
	else if (dataFile.find("Thrust") != string::npos || dataFile.find("C") 
		 != string::npos)  {
	  dataFormat = 8;
	}
	else if (dataFile.find("Nch") != string::npos )  {
	  dataFormat = 1;
	}
	else if (dataFile.find("LogInvx") != string::npos )  {
	  dataFormat = 7;
	}
      } 
      // L3 Data
      else if (dataFile.find("L3") != string::npos) {
	dataFormat = 3;
	if (dataFile.find("91-Nch-udsc") != string::npos) {
	  dataFormat = 2;
	} else if (dataFile.find("91-Nch-b") != string::npos) {
	  dataFormat = 1;
	} else if (dataFile.find("91-BWide-udsc") != string::npos || 
		   dataFile.find("91-BTot-udsc") != string::npos) {
	  dataFormat = 9;
	} 
      } 
      // Opal Data
      else if (dataFile.find("Opal") != string::npos) {
	dataFormat = 3;
      }
      // SLD Data
      else if (dataFile.find("SLD") != string::npos) {
	dataFormat = 10;
      }
      // General Data
      else if (dataFile.find("LnY") != string::npos) {
	dataFormat = 5;
      } 
      else if (dataFile.find("mesons") != string::npos || dataFile.find("baryons") != string::npos) {
	dataFormat = 0;
      }
	
      // Read next data file line
      while (1) {
	// mcplots format
	if (dataFormat == 11 || dataFormat == 12) {
	  if (dataFormat == 11) in >> xlo >> xc >> xhi >> yc >> sig1 >> sig2;
	  else in >> xlo >> xc >> xhi >> yc >> sig1 >> sig2 >> sig3 >> sig4;
	  if (!in.good()) break;
	  // Process x value
	  x[nlines] = xc;
	  exL[nlines] = xc-xlo;
	  exR[nlines] = xhi-xc;
	  // Compute integral 
	  sumsig += yc;
	  // Process y value
	  y[nlines] = yc;
	  eyStatH[nlines] = sig1;
	  eyStatL[nlines] = sig2;
	  eyTotH[nlines] = sig1+sig3;
	  eyTotL[nlines] = sig2+sig4;
	}
	else if (dataFormat == 0 || dataFormat == 1) {
	  if (dataFormat == 0) {
	    sig2 = 0.0;
	    in >> xc >> yc >> sig1; 
	  } else if (dataFormat == 1) {
	    in >> xc >> yc >> sig1 >> sig2;	  
	  }
	  if (!in.good()) break;
	  // Process x value
	  x[nlines] = xc;
	  if (nlines == 0) {
	    exL[nlines] = 1;
	    exR[nlines] = 1;
	    if (dataFormat == 0) exL[nlines] = 0.5;
	    if (dataFormat == 0) exR[nlines] = 0.5;
	    // If the first bin has a greater-than-zero central value and
	    // 0.5 or 1.0 makes it begin at < 0, assume it just begins at zero
	    if (xc > 0 && xc - exL[nlines] < 0) {
	      exL[nlines] = xc;
	      exR[nlines] = xc;
	    }
	  } else {
	    // For further bins, assume width of previous bin is at most half
	    // 
	    if (exR[nlines-1] > (xc-x[nlines-1])/2.) {
	      exR[nlines-1] = (xc-x[nlines-1])/2.;
	      exL[nlines-1] = exR[nlines-1];
	    }
	    exL[nlines] = xc - (x[nlines-1]+exR[nlines-1]);
	    exR[nlines] = exL[nlines];
	  }
	  // Compute integral 
	  sumsig += yc;
	  // Process y value
	  y[nlines] = yc;
	  eyStatL[nlines] = sig1;
	  eyStatH[nlines] = sig1;
	  eyTotL[nlines] = sig1+sig2;
	  eyTotH[nlines] = sig1+sig2;
	} 
	// Format 2
	else if (dataFormat == 2) {
	  in >> xlo >> xhi >> yc >> sig1 >> sig2 >> sig3 >> sig4;
	  if (!in.good()) break;
	  // Take abs values
	  sig1 = abs(sig1);
	  sig2 = abs(sig2);
	  sig3 = abs(sig3);
	  sig4 = abs(sig4);
	  // Process x value
	  x[nlines] = (xlo+xhi)/2.;
	  exL[nlines] = 1.0;
	  exR[nlines] = 1.0;
	  xlo = x[nlines] - exL[nlines];
	  xhi = x[nlines] + exR[nlines];
	  // Process y value
	  y[nlines] = yc;
	  eyStatH[nlines] = sig1;
	  eyStatL[nlines] = sig2;
	  // Add stat and sys linearly
	  eyTotH[nlines] = sig1 + sig3;
	  eyTotL[nlines] = sig2 + sig4;
	  sumsig += yc;
	}
	// Format 3
	else if (dataFormat == 3 || dataFormat == 4 || abs(dataFormat) == 7 || dataFormat == 8 || dataFormat == 9 || dataFormat == 10) {
	  if (dataFormat == 3) {
	    in >> xlo >> xhi >> yc >> sig1 >> sig2 >> sig3 >> sig4;
	  } else if (dataFormat == 4) {
	    in >> xlo >> xhi >> yc >> sig1 >> sig3 >> sig4;
	  } else if (abs(dataFormat) == 7) {
	    in >> xlo >> xhi >> yc >> sig1 >> sig3;
	  } else if (abs(dataFormat) == 8) {
	    in >> xlo >> xhi >> yc >> sig1 >> sig2 >> sig3 >> sig4 >> sig5 >> sig6;
	  } else if (abs(dataFormat) == 9) {
	    in >> xc >> xlo >> xhi >> yc >> sig1 >> sig2 >> sig3 >> sig4 ;
	  } else if (abs(dataFormat) == 10) {
	    in >> xc >> xlo >> xhi >> yc >> sig1 >> sig2 >> sig3 >> sig4 >> sig5 >> sig6;
	  }
	  if (!in.good()) break;
	  if (dataFormat == 4) sig2 = sig1;
	  if (abs(dataFormat) == 7) {
	    sig2 = sig1;
	    sig4 = sig3;
	  } 
	  else if (dataFormat == 8 || dataFormat == 10) {
	    sig3 = sqrt(pow2(sig3)+pow2(sig5));
	    sig4 = sqrt(pow2(sig4)+pow2(sig6));
	  }
	  if (dataFormat == -7) {
	    xlo  = log(xlo);
	    xhi  = log(xhi);
	  }	  
	  // Take abs values
	  sig1 = abs(sig1);
	  sig2 = abs(sig2);
	  sig3 = abs(sig3);
	  sig4 = abs(sig4);
	  // Process x value
	  x[nlines] = (xlo+xhi)/2.;
	  if (dataFormat == 9 || dataFormat == 10) x[nlines] = xc;
	  exL[nlines] = (xhi-xlo)/2.;
	  exR[nlines] = (xhi-xlo)/2.;
	  if (dataFormat == 9 || dataFormat == 10) {
	    exL[nlines] = (xc-xlo);
	    exR[nlines] = (xhi-xc);
	  }
	  xlo = x[nlines] - exL[nlines];
	  xhi = x[nlines] + exR[nlines];
	  // Process y value
	  y[nlines] = yc;
	  eyStatH[nlines] = sig1;
	  eyStatL[nlines] = sig2;
	  // Add stat and sys linearly
	  eyTotH[nlines] = sig1 + sig3;
	  eyTotL[nlines] = sig2 + sig4;
	  sumsig += yc * (xhi - xlo);
	  //	  cout << dataFile<<" : "<< xlo << " "<<xhi<<endl;
	}
	// Format 5
	else if (dataFormat == 5 || dataFormat == 6) {
	  in >> xlo >> xhi >> yc >> sig1 >> sig2 >> sig3;
	  if (!in.good()) break;
	  // Take abs values
	  sig1 = abs(sig1);
	  sig2 = abs(sig2);
	  sig3 = abs(sig3);
	  // Process x value
	  x[nlines] = (xlo+xhi)/2.;
	  exL[nlines] = (xhi-xlo)/2.;
	  exR[nlines] = (xhi-xlo)/2.;
	  xlo = x[nlines] - exL[nlines];
	  xhi = x[nlines] + exR[nlines];
	  // Process y value
	  y[nlines] = yc;
	  if (dataFormat == 5) {
	    eyStatH[nlines] = sig1;
	    eyStatL[nlines] = sig2;
	  } else {
	    eyStatH[nlines] = sig2;
	    eyStatL[nlines] = sig2;
	  }
	  // Add stat and sys linearly
	  if (dataFormat == 5) {
	    eyTotH[nlines] = sig1 + sig2 + sig3;
	    eyTotL[nlines] = sig2 + sig2 + sig3;
	  } else {
	    eyTotH[nlines] = sig1;
	    eyTotL[nlines] = sig1;
	  }
	  sumsig += yc * (xhi - xlo);
	  //	  cout << dataFile<<" : "<< xlo << " "<<xhi<<endl;
	}
	else {
	  cout<<"(VinciaRoot:) Unrecognized file format"<<endl;
	  return false;
	}
	nlines++;
	//	if (nlines < 200) printf("xlo=%8f, xhi=%8f, y=%8f, ey=+%8f-%8f\n",xlo,xhi,yc,abs(sig1)+abs(sig3),abs(sig2)+abs(sig4));
      }
      if (verbose >= 1) {
	printOut("VinciaRoot",num2str(nlines)+" readable data points in "+dataFile);
      }	
      
      // Read SRC and REF if present
      vDataSrc[iHist] = " ";
      vDataRef[iHist] = " ";
      vDataEcm[iHist] = " ";
      in.clear();
      char line[256];
      while (1) {
	in.getline(line,256);
	if (in.good()) {
	  string sline(line);		    
	  int ipad = 1;
	  if (sline.substr(sline.find("=")+1,1) == " ") ipad = 2;
	  if (sline.find("SRC") != string::npos) {
	    vDataSrc[iHist] = sline.substr(sline.find("=")+ipad,sline.size());
	  } else if (sline.find("REF") != string::npos) {
	    vDataRef[iHist] = sline.substr(sline.find("=")+ipad,sline.size());
	  } else if (sline.find("SQRT(S)") != string::npos) {
	    vDataEcm[iHist] = sline.substr(sline.find(":")+ipad,sline.size());
	  }
	} else {
	  break;
	}
      }

      // Return if no usable data points could be read
      if (nlines <= 0) {
	in.close();
	return false;
      }

      // Normalize?
      if (normFactor != 1.0) {
	for (int i=0; i<200; i++) {
	  y[i] *= normFactor;
	  eyTotL[i] *= normFactor;
	  eyTotH[i] *= normFactor;
	  eyStatL[i] *= normFactor;
	  eyStatH[i] *= normFactor;
	}
      }

      if (normalizeData && sumsig >= 0.0) {
	for (int i=0; i<200; i++) {
	  y[i] /= sumsig;
	  eyTotL[i] /= sumsig;
	  eyTotH[i] /= sumsig;
	  eyStatL[i] /= sumsig;
	  eyStatH[i] /= sumsig;
	}
      }

      // Special case for L3 and Alpeh Thrust measurements: T -> 1-T
      if (dataFile == "data/L3-91-Thrust-udsc.vec" || dataFile == "data/L3-91-Thrust-b.vec" || dataFile == "data/Aleph-91-Thrust.vec") {
	// Invert
	for (int i=0; i<nlines; i++) {
	  xT[i]       = 1-x[nlines-i-1];
	  exLT[i]     = exR[nlines-i-1];
	  exRT[i]     = exL[nlines-i-1];
	  yT[i]       = y[nlines-i-1];
	  eyTotLT[i]  = eyTotL[nlines-i-1];
	  eyTotHT[i]  = eyTotH[nlines-i-1];
	  eyStatLT[i] = eyStatL[nlines-i-1];
	  eyStatHT[i] = eyStatH[nlines-i-1];
	}
	// Save 
	for (int i=0; i<nlines; i++) {
	  x[i]        = xT[i];
	  exL[i]      = exRT[i];
	  exR[i]      = exLT[i];
	  y[i]        = yT[i];
	  eyTotL[i]   = eyTotLT[i];
	  eyTotH[i]   = eyTotHT[i];
	  eyStatL[i]  = eyStatLT[i];
	  eyStatH[i]  = eyStatHT[i];
	}
      }

      // Graph for error sum 
      TGraphAsymmErrors grDataTot(nlines,x,y,exL,exR,eyTotL,eyTotH);
      grDataTot.SetMarkerColor(dataColor);
      grDataTot.SetMarkerStyle(21);
      grDataTot.SetLineColor(dataColor);
      grDataTot.SetMarkerSize(0.5);
      grDataTot.SetName("DataTot");


      // Graph for stat component only
      TGraphAsymmErrors grDataStat(nlines,x,y,exL,exR,eyStatL,eyStatH);
      grDataStat.SetMarkerColor(dataColor);      
      grDataStat.SetMarkerSize(0.5);
      grDataStat.SetLineColor(dataColor);
      grDataStat.SetMarkerStyle(21);
      grDataStat.SetName("DataStat");

      // Add this data set
      vGrDataTot[iHist]=grDataTot;
      vGrDataStat[iHist]=grDataStat;


      // Create normalized versions (normalized reference)
      for (int i=0; i<200; i++) {
	if (y[i] != 0.0 && i < nlines) {
	  eyStatL[i] /= y[i];
	  eyStatH[i] /= y[i];
	  eyTotL[i]  /= y[i];
	  eyTotH[i]  /= y[i];
	  // Bound the error by 50% to avoid plotting outside plot
	  // (got to be a smarter way of doing this ...)
	  eyStatL[i] = min(Float_t(0.5),eyStatL[i]);
	  eyStatH[i] = min(Float_t(0.5),eyStatH[i]);
	  eyTotL[i]  = min(Float_t(0.5),eyTotL[i]);
	  eyTotH[i]  = min(Float_t(0.5),eyTotH[i]);
	  y[i]       = 1.0;
	} else {
	  y[i] = 1.0;
	  eyStatL[i] = 0.5;
	  eyStatH[i] = 0.5;
	  eyTotL[i] = 0.5;
	  eyTotH[i] = 0.5;
	}
      }   
      TGraphAsymmErrors grDataTotN(nlines,x,y,exL,exR,eyTotL,eyTotH);      
      grDataTotN.SetMarkerColor(1);
      grDataTotN.SetMarkerStyle(21);
      grDataTotN.SetLineColor(1);
      grDataTotN.SetMarkerSize(0.5);
      grDataTotN.SetName("DataTotN");

      TGraphAsymmErrors grDataStatN(nlines,x,y,exL,exR,eyStatL,eyStatH);
      grDataStatN.SetMarkerColor(1);      
      grDataStatN.SetMarkerSize(0.3);            
      grDataStatN.SetLineColor(1);
      grDataStatN.SetMarkerStyle(21);
      grDataStatN.SetName("DataStatN");

      // Add this normalized data set
      vGrDataTotN[iHist]=grDataTotN;
      vGrDataStatN[iHist]=grDataStatN;

      // Add placeholders for theory normalized to data
      TGraphAsymmErrors* grTheoryN
	= new TGraphAsymmErrors(nlines,x,zero,exL,exR,zero,zero);
      grTheoryN->SetMaximum(1.5);
      grTheoryN->SetMinimum(0.5);
      grTheoryN->SetName("TheoryN");

      TGraphAsymmErrors* grTheoryVarN 
	= new TGraphAsymmErrors(nlines,x,zero,exL,exR,zero,zero);
      grTheoryVarN->SetMaximum(1.5);
      grTheoryVarN->SetMinimum(0.5);
      grTheoryVarN->SetName("TheoryVarN");

      // Store these
      vGrTheoryN[iHist]=grTheoryN;
      vGrTheoryVarN[iHist]=grTheoryVarN;

      // Add placeholder for a normalized version of a reference curve
      TGraphAsymmErrors* grRefN
	= new TGraphAsymmErrors(nlines,x,zero,exL,exR,zero,zero);
      vGrRefStatN[iHist] = grRefN;

      return true;

      // Close stream
      in.close();

      // Return succes
      return true;
    }
    

    //------------------------------------------------------------------------
    // addRefFile
    // Read file containing second comparison reference and store points
    // Format must be    xlo  xc   xhi    y    +dy   -dy

    bool addRefFile(string refFile, bool normalize=true) {

      // Define read-in stream
      ifstream in;

      // Index
      int iHist = vTheory.size()-1;

      // Return false if fails
      in.open(refFile.c_str());
      if (! in) return false;

      // Arrays for read-in storage
      Float_t xc,xlo,xhi,yc,sig1,sig2;
      Int_t nlines = 0;
      Float_t x[200],exL[200],exR[200];
      Float_t y[200],eyStatL[200],eyStatH[200];
      Float_t sumsig=0.0;
      
      // Set data format depending on which file we're reading
      double normFactor = vTheory[iHist]->GetNormFactor();
      if (normFactor <= 0.0) normFactor = 1.0;

      // Read next data file line
      while (1) {
	in >> xlo >> xc >> xhi >> yc >> sig1 >> sig2;
	if (!in.good()) break;
	// Process x value
	x[nlines] = xc;
	exL[nlines] = xc - xlo;
	exR[nlines] = xhi - xc;
	// Compute integral 
	sumsig += yc * (xhi - xlo);
	// Process y value
	y[nlines] = yc;
	eyStatH[nlines] = abs(sig1);
	eyStatL[nlines] = abs(sig2);
	nlines++;
      }
      if (verbose >= 2) {
	printOut("VinciaRoot",num2str(nlines)+" readable ref points in "+refFile);
      }	
      
      // Read SRC
      vRefSrc[iHist] = " ";
      in.clear();
      char line[256];
      while (1) {
	in.getline(line,256);
	if (in.good()) {
	  string sline(line);		    
	  int ipad = 1;
	  if (sline.substr(sline.find("=")+1,1) == " ") ipad = 2;
	  if (sline.find("SRC") != string::npos) {
	    vRefSrc[iHist] = sline.substr(sline.find("=")+ipad,sline.size());
	  }
	} else {
	  break;
	}
      }

      // Return if no usable data points could be read
      if (nlines <= 0) {
	in.close();
	return false;
      }
      
      // Check for special data set with double-size bins
      // (E.g., Nch with only even values)
      if (vData[iHist].find("Nch") != string::npos) normFactor *= 2;

      // Normalize?
      if (normFactor != 1.0) {
	for (int i=0; i<200; i++) {
	  y[i] *= normFactor;
	  eyStatL[i] *= normFactor;
	  eyStatH[i] *= normFactor;
	}
      }

      if (normalize && sumsig >= 0.0) {
	for (int i=0; i<200; i++) {
	  y[i] /= sumsig;
	  eyStatL[i] /= sumsig;
	  eyStatH[i] /= sumsig;
	}
      }
      

      // Graph for stat component only
      TGraphAsymmErrors grRefStat(nlines,x,y,exL,exR,eyStatL,eyStatH);
      grRefStat.SetMarkerColor(refColor);      
      grRefStat.SetMarkerSize(0.6);
      grRefStat.SetLineColor(refColor);
      grRefStat.SetLineStyle(1);
      grRefStat.SetLineWidth(2);
      grRefStat.SetMarkerStyle(24);
      grRefStat.SetName("RefStat");

      // Add this reference plot
      vGrRefStat[iHist]=grRefStat;

      // Theory / Data ?
      if (vData[iHist] != "none") {

	// Now normalize theory (and ref) to data and plot	
	TGraphAsymmErrors* dataStat = &vGrDataStat[iHist];
	TGraphAsymmErrors* ref      = &vGrRefStat[iHist];
	TGraphAsymmErrors* refN     = vGrRefStatN[iHist];
	int nPoints = dataStat->GetN();
	
	// Loop over data points 
	for (int iDat=0; iDat<nPoints; iDat++) {
	  Double_t xDat, xLoDat, xHiDat, yDat;
	  dataStat->GetPoint(iDat,xDat,yDat);
	  if (yDat == 0.0) continue;
	  xLoDat = xDat - dataStat->GetErrorXlow(iDat); 
	  xHiDat = xDat + dataStat->GetErrorXhigh(iDat); 

	  // Loop over ref points and rebin like data
	  double sumw     = 0.0;
	  double sumy     = 0.0;
	  // Statistical MC generation errors
	  double sumErr2  = 0.0;
	  int nPointsRef = ref->GetN();

	  for (int iRef=0; iRef<nPointsRef; iRef++){
	    Double_t xRef, xLoRef, xHiRef, yRef;
	    ref->GetPoint(iRef,xRef,yRef);
	    xLoRef = xRef - ref->GetErrorXlow(iRef); 
	    xHiRef = xRef + ref->GetErrorXhigh(iRef); 
	    if (xHiRef <= xLoDat) continue;
	    if (xLoRef > xHiDat) continue;
	    double bw = xHiRef-xLoRef;
	    double w = 1.0;
	    w -= max(0.,xLoDat-xLoRef)/bw;
	    w -= max(0.,xHiRef-xHiDat)/bw;
	    w = max(w,0.0);
	    sumw += w;	    
	    double yErr    = ref->GetErrorYlow(iRef);
	    sumy     += w*yRef;	    
	    sumErr2  += w*pow2(yErr);
	  }	  	  
	  // Sanity check
	  if (sumw <= 0.) continue;

	  // Compute average values and total stat error
	  double yRef   = sumy/sumw;	  
	  double eyRefStat  = sqrt(sumErr2)/sumw;

	  // Normalize by data
	  Double_t yRefN     = yRef/yDat;
	  Double_t eyRefN    = eyRefStat/yDat;	  

	  // Make sure error bars don't go beyond edge of plot
	  refN->SetPoint(iDat,xDat,yRefN);
	  refN->SetPointEYhigh(iDat,eyRefN);
	  refN->SetPointEYlow(iDat,eyRefN);
	  // Remove X error bars
	  refN->SetPointEXhigh(iDat,0.);
	  refN->SetPointEXlow(iDat,0.);

	}
      }

      // Return
      return true;

    }

    //------------------------------------------------------------------------
    // addTH1FPtr
    // Overloaded wrapper to add a single histogram to VinciaRoot
    // Main routine below.

    bool addTH1FPtr(TH1F* histPtrIn, string displayNameIn="none",
		    bool logy=false,bool normalize=true,bool showStatErr=false,
		    string dataFile="none", string refFile="none") {
      vector<TH1F*> tmp;
      tmp.push_back(histPtrIn);
      return addTH1FPtr(tmp, displayNameIn,logy,normalize,showStatErr,
			dataFile,refFile);
    }

    //------------------------------------------------------------------------
    // addTH1FPtr
    // Main routine to add a histogram or vector of histograms to VinciaRoot

    bool addTH1FPtr(vector<TH1F*> hisvPtrIn, string displayNameIn="none",
		    bool logy=false,bool normalize=true,bool showStatErr=false,
		    string dataFile="none", string refFile = "none") {

      // Sanity check
      if (hisvPtrIn.size() <= 0) return false;
      TH1F* histPtrIn = hisvPtrIn[0];

      // Add histogram to specified display, if exists
      vDisplayIndex.push_back(-1);
      bool displayExists = false;
      if (displayNameIn != "none") {
	for (int iDisplay=0; iDisplay<int(displayName.size()); iDisplay++) {
	  if (displayName[iDisplay] == displayNameIn) {
	    vDisplayIndex[vDisplayIndex.size()-1]=iDisplay;
	    displayExists = true;
	    break;
	  }
	}
      }

      //--------------------------------------------------------------
      // Add to list of plots
      vTheory.push_back(histPtrIn);
      vTheoryScale.push_back(1.0);
      int iHist = vTheory.size() - 1;

      //--------------------------------------------------------------
      // Set main draw option(s) for this histogram
      vNorm.push_back(normalize);
      vStat.push_back(showStatErr);
      vLogY.push_back(logy);      
      vInitPad.push_back(false);

      //--------------------------------------------------------------
      // User Colors (start with default blue color set)
      theoryMarkerColor.push_back(userColor1);
      theoryStatColor.push_back(userColor2);
      theoryUncColor.push_back(userColor3);
      theoryUncStatColor.push_back(userColor4);

      //--------------------------------------------------------------
      // Store vector of uncertainty histograms, if defined
      bool hasUnc = false;
      if (hisvPtrIn.size() > 1) {
	hasUnc = true;
	vTheoryVar[iHist] = hisvPtrIn;
	// Placeholder for relative uncertainties 
	vector<TGraph> grTheoryVar;
	vector<TH1F*> vThisTheoryVar = vTheoryVar[iHist];
	for (unsigned int iVar = 0; iVar < hisvPtrIn.size(); iVar++)   
	  grTheoryVar.push_back(TGraph(vThisTheoryVar[iVar]));
	vGrTheoryVar[iHist] = grTheoryVar;
	if (grTheoryVar.size() <= 1) hasUnc = false;
      }
      hasVar.push_back(hasUnc);

      //--------------------------------------------------------------
      // Add local copy that we can modify for plotting without bothering user
      TH1F hIn(*histPtrIn);
      double normFactor = hIn.GetNormFactor();
      if (normFactor <= 0.0) normFactor = 1.0;
      hIn.SetMinimum(hIn.GetMinimum()/normFactor);
      hIn.SetMaximum(hIn.GetMaximum()/normFactor);
      hIn.GetXaxis()->SetLabelFont(42);
      hIn.GetXaxis()->SetLabelSize(0.05);
      hIn.GetXaxis()->SetNdivisions(505);
      hIn.GetXaxis()->SetLabelOffset(0.02);
      hIn.GetYaxis()->SetLabelFont(42);
      hIn.GetYaxis()->SetLabelSize(0.05);
      hIn.GetYaxis()->SetNdivisions(606);
      hIn.GetYaxis()->SetLabelOffset(0.02);      
      hIn.SetTickLength(0.05,"X");
      hIn.SetTitle("");
      if (string(hIn.GetTitle()).find("Mesons") != string::npos || 
	  string(hIn.GetTitle()).find("Baryons") != string::npos) 
	hIn.GetXaxis()->SetLabelSize(0.07);  
      vAxis.push_back(hIn);
      vThCopyLine.push_back(hIn);
      vThCopyVar.push_back(hIn);
      vThCopyVarUnc.push_back(hIn);
      
      //--------------------------------------------------------------
      // Data and Theory / Data 
      if (dataFile != "none") {
	// Check if data file exists
	if (!addDataFile(dataFile,normalize)) dataFile="none";
      }
      vData.push_back(dataFile);
      // Add axis for normalized histograms, with zero contents
      if (dataFile != "none") {
	TH1F axis(hIn);      
	axis.Reset();
	axis.SetStats(false);
	axis.SetTitle("");
	axis.SetMinimum(0.5);
	axis.SetMaximum(1.5);
	vAxisN[iHist]=axis;
      }

      //--------------------------------------------------------------
      // Theory uncertainties : make axis for relative uncertainties
      if (hasUnc) {
	TH1F axis(hIn);
	axis.Reset();
	axis.SetStats(false);
	axis.SetTitle("");
	axis.SetMinimum(-0.7);
	axis.SetMaximum(1.15);
	vAxisU[iHist]=axis;
      }

      //--------------------------------------------------------------
      // Comparison to other data or MC file
      if (refFile != "none") {
	// Check if ref file exists
	if (!addRefFile(refFile,normalize)) refFile="none";
      }
      vRef.push_back(refFile);
      vTit.push_back(new TLatex(0.0,0.0," "));
 
      // Always return true
      return true;
    }

    // Update display
    // iDisplayIn
    //    == -1 : update all
    //    >=  0 : update specific display
    bool updateDisplay(int iDisplayIn=-1) {      

      // Do nothing if no histograms
      if (vTheory.size() <= 0) return false;

      // Loop over displays
      for (int iDisplay=0; iDisplay<int(displays.size()); iDisplay++) {

	// If asking for specific display, only update that display
	if (iDisplayIn >= 0 && iDisplay != iDisplayIn) continue;

	// Limit number of plots on each display to finite resolution
	int nPixMin = 10000;
	unsigned int nPlotMax = max(xSiz[iDisplay]*ySiz[iDisplay]/nPixMin,1);
	double aspectRatio = (1.*xSiz[iDisplay])/ySiz[iDisplay];
	unsigned int nPlots=0;
	for (int iPlot=0; iPlot<int(vDisplayIndex.size()); iPlot++) {
	  if (vDisplayIndex[iPlot] == iDisplay) nPlots++;
	}
	nPlots = min(nPlotMax,nPlots);	
	// Always at least 1 to avoid divide-by-zero
	if (nPlots <= 1) nPlots = 1;

	// Compute necessary number of rows and columns
	unsigned int nCols = 1;
	unsigned int nRows = 1;
	unsigned int nTot  = 1;
	// Special for 4 plots in 4x1
	if (nPlots == 4) {
	  nTot  = 4;
	  if (aspectRatio > 1.6) {
	    nCols = 4;
	    nRows = 1;
	  } else {
	    nCols = 2;
	    nRows = 2;
	  }
	}
	else if (nPlots == 5 || nPlots == 6) {
	  nTot = 6;
	  if (aspectRatio >= 1.0) {
	    nCols = 3;
	    nRows = 2;
	  } else {
	    nCols = 2;
	    nRows = 3;
	  }
	}
	while (nTot < nPlots) {
	  if (nCols < 3*nRows) {
	    nCols += 1;
	  } else {
	    nRows += 1;
	  }
	  nTot   = nCols * nRows;
	}
	displays[iDisplay]->cd();	
	// Check if new division necessary
	if (nRows != nRowsSav[iDisplay] || nCols != nColsSav[iDisplay]) {
	  // Divide and update old
	  displays[iDisplay]->Divide(nCols,nRows);
	  nColsSav[iDisplay] = nCols;
	  nRowsSav[iDisplay] = nRows;	  
	  double padAspect = aspectRatio / nCols * nRows;
	  // Check which ones need to be subdivided into data norm
	  int iPlot=-1;
	  for (long unsigned int iHist=0; iHist<vTheory.size(); iHist++) {
	    if (vDisplayIndex[iHist] != iDisplay) continue;
	    else iPlot++;
	    int nDiv = 1;
	    if (vData[iHist] != "none") nDiv++;
	    if (hasVar[iHist] && uBands.size() > 1) nDiv++;
	    // Set margins
	    double topMargin = 0.1;
	    double botMargin = 0.15;
	    if (padAspect > 1.1) {
	      topMargin = 0.05;
	      botMargin = 0.20;
	    }
	    if (nDiv >= 2) botMargin += 0.12;
	    // Set pads
	    if (nDiv == 1) {
	      displays[iDisplay]->cd(iPlot+1);
	      gPad->SetTopMargin(topMargin);
	      gPad->SetBottomMargin(botMargin);
	      gPad->SetLeftMargin(0.13);
	      gPad->SetRightMargin(0.05);
	    } else if (nDiv == 2) {
	      displays[iDisplay]->cd(iPlot+1);
	      gPad->Divide(1,2);
	      gPad->cd(1);
	      gPad->SetPad(0.0,0.38,1.0,1.0);
	      gPad->SetTopMargin(topMargin);
	      gPad->SetBottomMargin(0.0);
	      gPad->SetLeftMargin(0.16);
	      gPad->SetRightMargin(0.05);
	      if (! vInitPad[iHist]) {
		gPad->Draw();
	      } else {
		gPad->Modified();
	      }
	      displays[iDisplay]->cd(iPlot+1);
	      gPad->cd(2);
	      gPad->SetPad(0.0,0.0,1.0,0.38);
	      gPad->SetTopMargin(0.0);
	      gPad->SetBottomMargin(botMargin);
	      gPad->SetLeftMargin(0.16);
	      gPad->SetRightMargin(0.05);
	      if (! vInitPad[iHist]) {
		gPad->Draw();
	      } else {
		gPad->Modified();
	      }
	    }
	    else if (nDiv == 3) {
	      displays[iDisplay]->cd(iPlot+1);
	      gPad->Divide(1,3);
	      gPad->cd(1);
	      gPad->SetPad(0.0,0.47,1.0,1.0);
	      gPad->SetTopMargin(topMargin);
	      gPad->SetBottomMargin(0.0);
	      gPad->SetLeftMargin(0.16);
	      gPad->SetRightMargin(0.05);
	      if (! vInitPad[iHist]) {
		gPad->Draw();
	      } else {
		gPad->Modified();
	      }
	      displays[iDisplay]->cd(iPlot+1);
	      gPad->cd(2);
	      gPad->SetPad(0.0,0.35,1.0,0.47);
	      gPad->SetTopMargin(0.0);
	      gPad->SetBottomMargin(0.0);
	      gPad->SetLeftMargin(0.16);
	      gPad->SetRightMargin(0.05);
	      if (! vInitPad[iHist]) {
		gPad->Draw();
	      } else {
		gPad->Modified();
	      }
	      displays[iDisplay]->cd(iPlot+1);
	      gPad->cd(3);
	      gPad->SetPad(0.0,0.0,1.0,0.35);
	      gPad->SetTopMargin(0.0);
	      gPad->SetBottomMargin(botMargin);
	      gPad->SetLeftMargin(0.16);
	      gPad->SetRightMargin(0.05);
	      if (! vInitPad[iHist]) {
		gPad->Draw();
	      } else {
		gPad->Modified();
	      }
	    }
	  }
	}
	// Plot histos
	for (long unsigned int iHist=0; iHist<vTheory.size(); iHist++) {
	  if (vDisplayIndex[iHist] != iDisplay) continue;
	  updatePlot(iHist);
	}
	displays[iDisplay]->cd();
	displays[iDisplay]->Modified();
	displays[iDisplay]->Update();
      }
      return true;
    }

    //------------------------------------------------------------------------
    // updatePlot: 
    // Main routine to update a specific plot

    bool updatePlot(long unsigned int iHist=0) {

      // Sanity check
      if (iHist >= vDisplayIndex.size()) return false;

      // Goto display
      int iDisplay = vDisplayIndex[iHist];
      if (iDisplay == -1) return false;

      // Goto subdivision 
      unsigned int iPlot = -1;
      for (long unsigned int iHistTmp=0; iHistTmp<=iHist; iHistTmp++) {
	if (iHistTmp >= vTheory.size()) break;
	if (vDisplayIndex[iHistTmp] != iDisplay) continue;
	else iPlot++;
      }
      displays[iDisplay]->cd(iPlot+1);      

      // Get name
      string hName(vTheory[iHist]->GetTitle());

      // How many subdivisions?
      int nDiv = 1;
      if (vData[iHist] != "none") nDiv++;
      if (hasVar[iHist]) nDiv++;

      // If more than one subdivision, go to first one
      if (nDiv >= 2) {
	gPad->cd(1);
      }

      // Check for special data set with double-size bins
      // (E.g., Nch with only even values)
      double normFactor = 1.0;
      if (vData[iHist].find("Nch") != string::npos) normFactor *= 2;

      // If normalizing, create copy to work on
      double sum  = 1.0;
      // Find location of max bin (left or right half, for legend placing)
      Int_t iBinMax = 0;
      double valMax = vTheory[iHist]->GetBinContent(1);
      // Update copy
      // And change to 90% statistical uncertainties.     
      // Ignore underflow and overflow bins
      for (int iBin=1; iBin<=vTheory[iHist]->GetNbinsX() ; iBin++) {
	double val = vTheory[iHist]->GetBinContent(iBin);
	// Find max bin
	if (val > valMax) {
	  iBinMax = iBin;
	  valMax  = val;
	}
	// Now copy over contents and rescale stat uncertainty
	vAxis[iHist].SetBinContent(iBin,val);
	vAxis[iHist].SetBinError(iBin,1.645*vTheory[iHist]->GetBinError(iBin));
	vThCopyLine[iHist].SetBinContent(iBin,val);
	/* Skip empty bins in line copy?
	else if (iBin >= 2 && iBin < vTheory[iHist]->GetNbinsX()) {
	  double valinterpolate = 0.5 * (vTheory[iHist]->GetBinContent(iBin-1)
					 + vTheory[iHist]->GetBinContent(iBin+1));
	  vThCopyLine[iHist].SetBinContent(iBin,valinterpolate);
	}
	*/
	// Do uncertainty band variations
	if (hasVar[iHist]) {	  
	  double varHi    = val;
	  double varLo    = val;
	  double varUncHi = val
	    +1.645*vTheory[iHist]->GetBinError(iBin);
	  double varUncLo = val
	    -1.645*vTheory[iHist]->GetBinError(iBin);
	  vector<TH1F*> vVar = vTheoryVar[iHist];
	  vector<TGraph>& gVar = vGrTheoryVar[iHist];
	  // Uncertainty band envelope
	  for (int iVar=1; iVar<int(vVar.size()); iVar++) {
	    if (vVar[iVar]->GetEntries() > 0) {
	      varHi = max(varHi,vVar[iVar]->GetBinContent(iBin));
	      varLo = min(varLo,vVar[iVar]->GetBinContent(iBin));
	      varUncHi = max(varUncHi,vVar[iVar]->GetBinContent(iBin)
			     +1.645*vVar[iVar]->GetBinError(iBin));
	      varUncLo = min(varUncLo,vVar[iVar]->GetBinContent(iBin)
			     -1.645*vVar[iVar]->GetBinError(iBin));  	      
	    }
	  }
	  // Uncertainy bands: relative uncertainty plot
	  // (numbering excludes iBin = 0 (underflow)
	  double varAvg = 0.5*(varHi+varLo);	  
	  if (iBin > 0) {
	    for (unsigned int iBand=0; iBand < uBands.size(); iBand++) {
	      vector<int> vIndx = uBands[iBand].indices;
	      if (vIndx.size() <= 0) continue;
	      double thisVarHi = max(varLo,vVar[vIndx[0]]->GetBinContent(iBin));
	      double thisVarLo = thisVarHi;
	      for (unsigned int iIndx=1; iIndx<vIndx.size(); iIndx++) {
		int iVar = vIndx[iIndx];
		thisVarHi = max(thisVarHi,vVar[iVar]->GetBinContent(iBin));
		thisVarLo = min(thisVarLo,vVar[iVar]->GetBinContent(iBin));
	      }
	      double thisDelta  = thisVarHi - thisVarLo;
	      double thisRel = (varHi-varLo > 0.) ? thisDelta/(varHi-varLo) : 0.;
	      Double_t x, y;
	      if (iBand < gVar.size()) {
		gVar[iBand].GetPoint(iBin-1,x,y);
		gVar[iBand].SetPoint(iBin-1,x,thisRel);
	      }
	    }
	  }
	  vThCopyVar[iHist].SetBinContent(iBin,varAvg);
	  vThCopyVar[iHist].SetBinError(iBin,varHi-varAvg); 
	  vThCopyVarUnc[iHist].SetBinContent(iBin,0.5*(varUncHi+varUncLo));
	  vThCopyVarUnc[iHist].SetBinError(iBin,varUncHi-0.5*(varUncHi+varUncLo)); 
	} 
      }
      vAxis[iHist].SetEntries(vTheory[iHist]->GetEntries());

      if (vNorm[iHist]) {
	// Normalize Copies
	sum  = vAxis[iHist].GetSumOfWeights()*vAxis[iHist].GetBinWidth(1);
	if (normFactor != 1.0) sum /= normFactor;
	if (sum == 0.0) sum=1.0;
	vAxis[iHist].Scale(1.0/sum);
	vThCopyLine[iHist].Scale(1.0/sum);
	// Normalize uncertainty estimate by same as nominal
	// (assumes each band has sum(w) not much different from unity)
	vThCopyVar[iHist].Scale(1.0/sum);
	vThCopyVarUnc[iHist].Scale(1.0/sum);
      }

      // Set Y max and min values for plot       
      double YMax = vTheory[iHist]->GetMaximum();
      YMax = max(YMax,
		 1.05*vAxis[iHist].GetBinContent(vAxis[iHist].GetMaximumBin()));
      vAxis[iHist].SetMaximum(YMax);
      vThCopyLine[iHist].SetMaximum(YMax);
      vThCopyVar[iHist].SetMaximum(YMax);
      vThCopyVarUnc[iHist].SetMaximum(YMax);
      double YMin = vTheory[iHist]->GetMinimum();
      if (vLogY[iHist] && YMin <= 0.0) {
	YMin = max(1e-16,
		   0.95*vAxis[iHist].GetBinContent(vAxis[iHist].GetMinimumBin()));
      }
      vAxis[iHist].SetMinimum(YMin);
      vThCopyLine[iHist].SetMinimum(YMin);
      vThCopyVar[iHist].SetMinimum(YMin);
      vThCopyVarUnc[iHist].SetMinimum(YMin);
      // Draw histogram
      double facY(1.0);
      if (! vInitPad[iHist]) {
	// Size of plot
	double nPixPlotX = xSiz[iDisplay]/nColsSav[iDisplay];
	double nPixPlotY = ySiz[iDisplay]/nRowsSav[iDisplay];
	// Approximately remove left and right margins
	if (nDiv == 1) nPixPlotX *= (1.-0.13-0.05);
	else nPixPlotX *= (1.-0.16-0.05);
	// Approximately remove left and top margins
	if (nDiv == 1) nPixPlotY *= (1.-0.1-0.15);
	else nPixPlotY *= (1.-0.1);
	// -> size of first plot
	if (nDiv == 2) nPixPlotY *= 1.-0.38;
	else if (nDiv == 3) nPixPlotY *= 1.-0.47;
	double aspectRatio = nPixPlotX / nPixPlotY;
	// For very wide plots, increase font sizes
	if (aspectRatio > 1.4) facY *= sqrt(min(1.5,aspectRatio-0.4));
	// For very high plots, reduce Y font sizes
	if (aspectRatio < 1.1) facY *= sqrt(max(0.5,aspectRatio-0.1));

	// Set axis properties for main histogram plot
	if (vLogY[iHist]) gPad->SetLogy();
	// X axis 
	vAxis[iHist].GetXaxis()->SetLabelFont(42);
	vAxis[iHist].GetXaxis()->SetTitleFont(42);
	vAxis[iHist].GetXaxis()->SetLabelSize(0.06*facY);
	vAxis[iHist].GetXaxis()->SetLabelOffset(0.006*facY);
	vAxis[iHist].GetXaxis()->SetTitleSize(0.07*facY);
	vAxis[iHist].GetXaxis()->SetTitleOffset(1.05*facY);
	// Y axis
	vAxis[iHist].GetYaxis()->SetTitleFont(42);
	vAxis[iHist].GetYaxis()->SetTitleSize(0.075*facY);
	vAxis[iHist].GetYaxis()->SetTitleOffset(1.05/facY);
	vAxis[iHist].GetYaxis()->SetLabelFont(42);
	vAxis[iHist].GetYaxis()->SetLabelSize(0.07*facY);
	vAxis[iHist].GetYaxis()->SetLabelOffset(0.01/facY);
	if (nDiv == 1) {
	  vAxis[iHist].GetYaxis()->SetLabelSize(0.06*facY);
	  vAxis[iHist].GetYaxis()->SetTitleSize(0.07*facY);
	  vAxis[iHist].GetYaxis()->SetTitleOffset(1.05*facY);
	}
	// First: curve with user settings (will be replotted afterwards)
	vAxis[iHist].SetLineColor(theoryStatColor[iHist]);
	vAxis[iHist].SetFillColor(theoryStatColor[iHist]);	
	vAxis[iHist].SetMarkerColor(theoryMarkerColor[iHist]);	
	vAxis[iHist].SetMarkerStyle(20);	
	vAxis[iHist].SetMarkerSize(0.25);	
	vAxis[iHist].Draw("Pe");
	// Draw Systematic/Theory + Stat uncertainty
	if (hasVar[iHist]) {
	  vThCopyVarUnc[iHist].SetLineColor(theoryUncStatColor[iHist]);	
	  vThCopyVarUnc[iHist].SetMarkerColor(theoryUncStatColor[iHist]);	
	  vThCopyVarUnc[iHist].Draw("samee"); 	
	  // Draw Systematic/Theory uncertainty
	  vThCopyVar[iHist].SetFillColor(theoryUncColor[iHist]);
	  vThCopyVar[iHist].SetMarkerColor(theoryUncColor[iHist]);
	  vThCopyVar[iHist].SetLineColor(theoryUncColor[iHist]);
	  vThCopyVar[iHist].SetFillStyle(1001);
	  vThCopyVar[iHist].Draw("samee2"); 
	  // Overplot user plot to put in front
	  //vAxis[iHist].Draw("samee5"); 
	  vAxis[iHist].Draw("samePe"); 
	}
	// If secondary comparison plot, plot it
	if (vRef[iHist] != "none") {
	  vGrRefStat[iHist].Draw("sameX0lineP");	  
	}

	vThCopyLine[iHist].SetLineColor(theoryMarkerColor[iHist]);
	vThCopyLine[iHist].Draw("sameX0line");
	gPad->Update();

	// Set axis properties for Theory/Data plots
	if (vData[iHist] != "none") {
	  // X axis
	  vAxisN[iHist].GetXaxis()->SetTickLength(0.1);
	  vAxisN[iHist].GetXaxis()->SetLabelFont(42);
	  vAxisN[iHist].GetXaxis()->SetLabelSize(0.11*facY);
	  vAxisN[iHist].GetXaxis()->SetTitleFont(42);
	  vAxisN[iHist].GetXaxis()->SetTitleSize(0.125*facY);
	  vAxisN[iHist].GetXaxis()->SetTitleOffset(1.0);
	  // Y axis
	  vAxisN[iHist].GetYaxis()->SetLabelFont(42);
	  vAxisN[iHist].GetYaxis()->SetLabelSize(0.09*facY);
	  vAxisN[iHist].GetYaxis()->SetLabelOffset(0.01);
	  vAxisN[iHist].GetYaxis()->SetTitleFont(42);
	  vAxisN[iHist].GetYaxis()->SetTitleSize(0.0825*facY);
	  vAxisN[iHist].GetYaxis()->SetTitle("Theory/Data");	  
	  vAxisN[iHist].GetYaxis()->SetTitleOffset(0.825/facY);
	  vAxisN[iHist].GetYaxis()->CenterTitle();
	  vAxisN[iHist].GetYaxis()->SetTickLength(0.025);
	}
	// Set axis properties for relative uncertainties plots
	if (hasVar[iHist]) {
	  // X axis
	  vAxisU[iHist].GetXaxis()->SetTitleSize(0.5);
	  vAxisU[iHist].GetXaxis()->SetTitleFont(42);
	  vAxisU[iHist].GetXaxis()->SetLabelFont(42);
	  if (nDiv == 2) {
	    vAxisU[iHist].GetXaxis()->SetTitleOffset(1.0);
	    vAxisU[iHist].GetXaxis()->SetTitleSize(0.13*facY);
	    vAxisU[iHist].GetXaxis()->SetLabelSize(0.11*facY);
	    vAxisU[iHist].GetXaxis()->SetTickLength(0.11);	  
	  } else {
	    vAxisU[iHist].GetXaxis()->SetTitleOffset(1.1);
	    vAxisU[iHist].GetXaxis()->SetLabelSize(0.25);
	    vAxisU[iHist].GetXaxis()->SetTickLength(0.15);	  
	  }
	  // Y axis
	  vAxisU[iHist].GetYaxis()->SetNdivisions(3);
	  vAxisU[iHist].GetYaxis()->SetLabelFont(42);
	  vAxisU[iHist].GetYaxis()->SetLabelOffset(0.01);
	  vAxisU[iHist].GetYaxis()->SetTitle("Rel.Unc.");	  
	  vAxisU[iHist].GetYaxis()->CenterTitle();
	  vAxisU[iHist].GetYaxis()->SetTitleFont(42);
	  vAxisU[iHist].GetYaxis()->SetTitleOffset(0.26);
	  vAxisU[iHist].GetYaxis()->SetTickLength(0.02);
	  if (nDiv == 2) {
	    vAxisU[iHist].GetYaxis()->SetLabelSize(0.1);	    	  
	    vAxisU[iHist].GetYaxis()->SetTitleSize(0.1);
	    vAxisU[iHist].GetYaxis()->SetTitleOffset(0.75);
	  } else {
	    vAxisU[iHist].GetYaxis()->SetLabelSize(0.25);
	    vAxisU[iHist].GetYaxis()->SetTitleSize(0.26);
	    vAxisU[iHist].GetYaxis()->SetTitleOffset(0.26);
	  }
	}
	if (string(vTheory[iHist]->GetTitle()).find("Mesons") != string::npos || 
	    string(vTheory[iHist]->GetTitle()).find("Baryons") != string::npos) {
	  vAxis[iHist].GetXaxis()->SetLabelSize(0.18);  
	  vAxisN[iHist].GetXaxis()->SetLabelSize(0.18);
	}

	// Stats	
	if (optStat <= 0) 
	  vAxis[iHist].SetStats(0);
	else {
	  TPaveStats *st =
	    (TPaveStats*)vAxis[iHist].GetListOfFunctions()->FindObject("stats");
	  st->SetOptStat(optStat);
	  st->SetTextSizePixels(9);
	  st->SetTextFont(62);
	  double pSizX = 65.0/nPixPlotX;
	  double pSizY = 30.0/nPixPlotY;
	  pSizX=1.-min(pSizX,0.6);
	  pSizY=1.-min(pSizY,0.6);
	  st->SetX1NDC(pSizX);
	  st->SetX2NDC(0.99);
	  st->SetY2NDC(0.99);
	  st->SetY1NDC(pSizY);
	}

	// Draw legend
	if (mcLabel != " ") {
	  double xLeg = 0.65;
	  double xLegEnd = 0.95;
	  double yLeg = 0.61;
	  double yLegTop = 0.78;
	  //cout<<" iBinMax = "<<iBinMax<<" out of "<<vTheory[iHist]->GetNbinsX()<<
	  //  endl;
	  if (vData[iHist] != "none" && vRef[iHist] != "none") {
	    yLeg -= 0.08;
	  }
	  if (iBinMax > (vTheory[iHist]->GetNbinsX()+1)/2.) {
	    //cout<<vTheory[iHist]->GetTitle()<<" Left"<<" iBinMax = "<<iBinMax<<
	    //  " nBins = "<<vTheory[iHist]->GetNbinsX()+1<<endl;
	    xLeg -= 0.42;
	    xLegEnd -= 0.42;
	  }
	  TLegend *leg = new TLegend(xLeg,yLeg,xLegEnd,yLegTop,NULL,"rNDC");
	  leg->SetBorderSize(0);
	  leg->SetFillStyle(0);
	  if (vData[iHist] != "none") {
	    string dataLabel = "Data";
	    if (vDataSrc[iHist] != " ") dataLabel = vDataSrc[iHist];
	    TLegendEntry *entry=leg->AddEntry("gr",dataLabel.c_str(),"p");
	    entry->SetMarkerColor(dataColor);
	    entry->SetMarkerStyle(21);
	    entry->SetMarkerSize(0.5);
	    entry->SetTextAlign(12);
	    entry->SetTextSize(0.055*facY);
	    entry->SetTextFont(42);
	    entry->SetTextColor(1);
	  }
	  // Add uncertainty "band" on legend if showing one
	  if (hasVar[iHist]) {
	    TLegendEntry *entry=leg->AddEntry("gr",mcLabel.c_str(),"pl");
	    entry->SetLineColor(theoryMarkerColor[iHist]);
	    entry->SetMarkerColor(theoryUncColor[iHist]);
	    entry->SetMarkerSize(0.7);
	    entry->SetTextAlign(12);
	    entry->SetTextSize(0.055*facY);
	    entry->SetTextFont(42);
	    entry->SetTextColor(1);
	  } else {
	    TLegendEntry *entry=leg->AddEntry("gr",mcLabel.c_str(),"l");
	    entry->SetLineColor(theoryMarkerColor[iHist]);
	    entry->SetTextAlign(12);
	    entry->SetTextSize(0.055*facY);
	    entry->SetTextFont(42);
	    entry->SetTextColor(1);
	  }
	  if (vRef[iHist] != "none") {
	    string refLabel = "Ref";
	    if (vRefSrc[iHist] != " ") refLabel = vRefSrc[iHist];
	    TLegendEntry *entry=leg->AddEntry("gr",refLabel.c_str(),"pl");
	    entry->SetMarkerColor(refColor);
	    entry->SetLineColor(refColor);
	    entry->SetLineStyle(1);
	    entry->SetLineWidth(2);
	    entry->SetMarkerStyle(24);
	    entry->SetMarkerSize(0.6);
	    entry->SetTextAlign(12);
	    entry->SetTextSize(0.055*facY);
	    entry->SetTextFont(42);
	    entry->SetTextColor(1);
	  }
	  leg->Draw();
	}
      } else {
	gPad->Modified();
      }

      // Title
      string hisTitle = vTheory[iHist]->GetTitle();
      if (hisTitle != "") {
	int NBins = vTheory[iHist]->GetNbinsX();
	double xLo = vTheory[iHist]->GetBinLowEdge(1);
	double xHi = vTheory[iHist]->GetBinLowEdge(NBins+1);
	double xC = xLo + (xHi-xLo)/2.;
	double yMin = vTheory[iHist]->GetMinimum();
	double yMax = vTheory[iHist]->GetMaximum();
	double yC = yMin + 0.93*(yMax-yMin);
	if (vLogY[iHist]) {
	  double LnyC = log(yMin) + 0.93*log(yMax/yMin);
	  yC = exp(LnyC);
	}
	// Update Title location (needed if y axis scales)
	vTit[iHist]->SetText(xC,yC,hisTitle.c_str());
	// The rest only needed to be set first time
	if (  !vInitPad[iHist] ) {
	  vTit[iHist]->SetTextFont(62);
	  vTit[iHist]->SetTextAlign(22);
	  vTit[iHist]->SetTextColor(1);
	  vTit[iHist]->SetTextSize(0.065*facY);
	  vTit[iHist]->Draw();  
	}
	gPad->Modified();
      }

      // If Thrust, show dvisions between n-jet regions
      if (hName.find("1-T") != string::npos ) {
	// Draw line
	if ( !vInitPad[iHist] ) {
	  double t45 = 1. - 1./sqrt(3.);
	  double yMin = vTheory[iHist]->GetMinimum();
	  double yMax = vTheory[iHist]->GetMaximum();
	  lineThrust34 = TLine(0.3333,yMin,0.3333,yMax);
	  lineThrust34.SetLineStyle(3);
	  lineThrust34.SetLineColor(1);
	  lineThrust45 = TLine(t45,yMin,t45,yMax);
	  lineThrust45.SetLineStyle(3);
	  lineThrust45.SetLineColor(1);
	  double tLo = vTheory[iHist]->GetBinLowEdge(1);
	  double tHi = tLo + vTheory[iHist]->GetNbinsX() 
	    * vTheory[iHist]->GetBinWidth(1);
	  if (tLo < 0.3333 && tHi > 0.3333) lineThrust34.Draw();
	  if (tLo < t45 && tHi > t45) lineThrust45.Draw();
	}
      }

      // If C parameter, show dvisions between n-jet regions
      if (hName.find("C ") != string::npos ) {
	// Draw 3-jet line at 3/4
	if ( !vInitPad[iHist] ) {
	  double t34 = 0.75;
	  double yMin = vTheory[iHist]->GetMinimum();
	  double yMax = vTheory[iHist]->GetMaximum();
	  lineC34 = TLine(t34,yMin,t34,yMax);
	  lineC34.SetLineStyle(3);
	  lineC34.SetLineColor(1);
	  double tLo = vTheory[iHist]->GetBinLowEdge(1);
	  double tHi = tLo + vTheory[iHist]->GetNbinsX() 
	    * vTheory[iHist]->GetBinWidth(1);
	  if (tLo < t34 && tHi > t34) lineC34.Draw();
	}
      }

      // If Heavy Jet Mass, show dvisions between n-jet regions
      else  if (hName.find("Heavy Jet Mass") != string::npos ) {
	// Draw line
	if ( !vInitPad[iHist] ) {
	  double t34 = 1./3.;
	  double yMin = vTheory[iHist]->GetMinimum();
	  double yMax = vTheory[iHist]->GetMaximum();
	  lineMH34 = TLine(t34,yMin,t34,yMax);
	  lineMH34.SetLineStyle(3);
	  lineMH34.SetLineColor(1);
	  double tLo = vTheory[iHist]->GetBinLowEdge(1);
	  double tHi = tLo + vTheory[iHist]->GetNbinsX() 	    
	    * vTheory[iHist]->GetBinWidth(1);
	  if (tLo < t34 && tHi > t34) lineMH34.Draw();
	}
      }

      // Add data reference (if exists)
      if (vData[iHist] != "none" && vDataRef[iHist] != " " && !vInitPad[iHist]) {
	int NBins = vTheory[iHist]->GetNbinsX();
	double xLo = vTheory[iHist]->GetBinLowEdge(1);
	double xHi = vTheory[iHist]->GetBinLowEdge(NBins);
	double xC = xLo + (xHi-xLo)/2.;
	double yMin = vTheory[iHist]->GetMinimum();
	double yMax = vTheory[iHist]->GetMaximum();
	double yC = yMin + 0.08*(yMax-yMin)*facY;
	if (vLogY[iHist]) {
	  double LnyC = log(yMin) + 0.12*log(yMax/yMin);
	  yC = exp(LnyC);
	}
	string ref = "Data from "+vDataRef[iHist];
	TLatex* tex = new TLatex(xC,yC,ref.c_str());
	tex->SetTextFont(42);
	tex->SetTextAlign(22);
	tex->SetTextColor(12);
	tex->SetTextSize(0.04*facY);
	tex->Draw();  
      }

      // Add MC reference (if exists)
      if (mcRefs != " " && !vInitPad[iHist]) {
	int NBins = vTheory[iHist]->GetNbinsX();
	double xLo = vTheory[iHist]->GetBinLowEdge(1);
	double xHi = vTheory[iHist]->GetBinLowEdge(NBins);
	double xC = xLo + (xHi-xLo)/2.;
	double yMin = vTheory[iHist]->GetMinimum();
	double yMax = vTheory[iHist]->GetMaximum();
	double yC = yMin + 0.12*(yMax-yMin)*facY;
	if (vLogY[iHist]) {
	  double LnyC = log(yMin) + 0.18*log(yMax/yMin);
	  yC = exp(LnyC);
	}
	TLatex* texMC = new TLatex(xC,yC,mcRefs.c_str());
	texMC->SetTextFont(42);
	texMC->SetTextAlign(22);
	texMC->SetTextColor(12);
	texMC->SetTextSize(0.04*facY);
	texMC->Draw();  
      }

      //----------------------------------------------------------
      // If comparing to data, overplot data now
      if (vData[iHist] != "none") {
	// Start in first subdivision: theory vs data
	displays[iDisplay]->cd(iPlot+1);
	gPad->cd(1);	
	if (! vInitPad[iHist]) {
	  vGrDataTot[iHist].Draw("P");
	} else {
	  gPad->Modified();
	}
	// Now go to last subdivision: theory normalized to data
	displays[iDisplay]->cd(iPlot+1);
	gPad->cd(nDiv);	
	if (! vInitPad[iHist]) {
	  vAxisN[iHist].Draw();
	} else {
	  gPad->Modified();
	}

	// Now normalize theory (and ref) to data and plot	
	TGraphAsymmErrors* dataStat = &vGrDataStat[iHist];
	TGraphAsymmErrors* theoryN  = vGrTheoryN[iHist];
	TGraphAsymmErrors* theoryVarN = vGrTheoryVarN[iHist];
	int nPoints = dataStat->GetN();
	for (int iDat=0; iDat<nPoints; iDat++) {
	  Double_t xDat, xLoDat, xHiDat, yDat;
	  dataStat->GetPoint(iDat,xDat,yDat);
	  if (yDat == 0.0) continue;
	  xLoDat = xDat - dataStat->GetErrorXlow(iDat); 
	  xHiDat = xDat + dataStat->GetErrorXhigh(iDat); 
	  // Loop over theory bins and rebin like data
	  double sumw     = 0.0;
	  double sumy     = 0.0;
	  // Statistical MC generation errors
	  double sumErr2  = 0.0;
	  // Note: up/down variation curves must be added linearly!
	  double sumUpper = 0.0;
	  double sumLower = 0.0;
	  TH1F* h=&vAxis[iHist];
	  TH1F* hU=&vThCopyVar[iHist];
	  int nBins = h->GetNbinsX();
	  /*
	  cout << "iDat = "<<iDat <<" xlo = "<<xLoDat<<" xhi = "<<xHiDat<<endl; 
	  */
	  for (int iBin=0; iBin<nBins; iBin++){
	    double xLo = h->GetBinLowEdge(iBin+1);
	    double xHi = xLo + h->GetBinWidth(iBin+1);
	    if (xHi < xLoDat) continue;
	    if (xLo > xHiDat) continue;
	    double bw = xHi-xLo;
	    double w = 1.0;
	    w -= max(0.,xLoDat-xLo)/bw;
	    w -= max(0.,xHi-xHiDat)/bw;
	    w = max(w,0.0);
	    sumw += w;
	    double y       = h->GetBinContent(iBin+1);
	    double yU      = hU->GetBinContent(iBin+1);
	    double yErr    = h->GetBinError(iBin+1);	    
	    double yUpper  = yU+hU->GetBinError(iBin+1);
	    double yLower  = yU-hU->GetBinError(iBin+1);
	    sumy     += w*y;	    
	    sumErr2  += w*pow2(yErr);
	    sumUpper += w*yUpper;
	    sumLower += w*yLower;
	  }	  	  
	  // Sanity check
	  if (sumw <= 0.) continue;

	  // Compute average values and total stat error
	  double yTh   = sumy/sumw;	  
	  double eyThStat  = sqrt(sumErr2)/sumw;
	  double yVarUpper = sumUpper/sumw;
	  double yVarLower = sumLower/sumw;
	  double yVarTh    = 0.5*(yVarUpper+yVarLower);
	  double eyVarTh   = yVarUpper-yVarTh;

	  // Normalize by data
	  Double_t yThN     = yTh/yDat;
	  Double_t eyThN    = eyThStat/yDat;	  
	  Double_t yVarThN  = yVarTh/yDat;
	  // Add stat error in quadrature for normalized total uncertainty
	  Double_t eyVarThN = sqrt(pow2(eyVarTh)+pow2(eyThStat))/yDat;

	  // Bound normalized uncertainties inside the ratio plot (+/- 50%)
	  Double_t eyThNHi    = min(1.5-yThN,eyThN);
	  Double_t eyThNLo    = eyThN;
	  if (yThN-eyThNLo < 0.5) eyThNLo = yThN-0.5;
	  Double_t eyVarThNHi = min(1.5-yVarThN,eyVarThN);
	  Double_t eyVarThNLo = eyVarThN;
	  if (yVarThN-eyVarThNLo < 0.5) eyVarThNLo = yVarThN-0.5;

	  // Make sure error bars don't go beyond edge of plot
	  theoryN->SetPoint(iDat,xDat,yThN);
	  theoryN->SetPointEYhigh(iDat,eyThNHi);
	  theoryN->SetPointEYlow(iDat,eyThNLo);
	  // Remove X error bars
	  theoryN->SetPointEXhigh(iDat,0.);
	  theoryN->SetPointEXlow(iDat,0.);
	  theoryVarN->SetPoint(iDat,xDat,yVarThN);
	  theoryVarN->SetPointEYhigh(iDat,eyVarThNHi);
	  theoryVarN->SetPointEYlow(iDat,eyVarThNLo);

	}


	// Plot data points with stat (inner) and tot (outer) uncertainties
	if (! vInitPad[iHist]) {
	  TGraphAsymmErrors& dataTotN = vGrDataTotN[iHist];
	  dataTotN.SetLineColor(dataSysColor);
	  dataTotN.SetFillColor(dataSysColor);
	  dataTotN.SetMarkerColor(1);
	  dataTotN.SetFillStyle(1001);
	  dataTotN.Draw("e2");	
	  TGraphAsymmErrors& dataStatN = vGrDataStatN[iHist];
	  dataStatN.SetLineColor(dataStatColor);
	  dataStatN.SetFillColor(dataStatColor);
	  dataStatN.SetMarkerColor(1);
	  dataStatN.SetFillStyle(1001);
	  dataStatN.Draw("e2");	
	} else {
	  gPad->Modified();
	}

	// Plot theory histograms normalized to data
	if (!vInitPad[iHist]) {
	  if (hasVar[iHist]) {
	    theoryVarN->SetLineColor(theoryUncColor[iHist]);
	    theoryVarN->SetFillColor(theoryUncColor[iHist]);
	    theoryVarN->SetFillStyle(3354);
	    theoryVarN->Draw("samee2");	
	  }
	} else {
	  gPad->Modified();
	}

	// If secondary comparison plot, plot it
	if (vRef[iHist] != "none") {
	  if (! vInitPad[iHist]) {
	    vGrRefStatN[iHist]->SetMarkerColor(refColor);
	    vGrRefStatN[iHist]->SetMarkerStyle(24);
	    vGrRefStatN[iHist]->SetMarkerSize(0.6);
	    vGrRefStatN[iHist]->SetLineColor(refColor);
	    vGrRefStatN[iHist]->SetLineWidth(2);
	    vGrRefStatN[iHist]->SetLineStyle(1);
	    vGrRefStatN[iHist]->Draw("samePeZline");	  
	  }
	  else {
	    gPad->Modified();
	  }
	}

	// Finally plot main theory MC comparison line
	if (!vInitPad[iHist]) {
	  string opt = "samePeZline";
	  theoryN->SetLineColor(theoryStatColor[iHist]);
	  theoryN->SetLineStyle(1);
	  theoryN->SetLineWidth(1);
	  theoryN->SetFillColor(theoryStatColor[iHist]);
	  theoryN->SetFillStyle(0);
	  theoryN->SetMarkerColor(theoryMarkerColor[iHist]);
	  theoryN->SetMarkerStyle(20);
	  theoryN->SetMarkerSize(0.5);
	  theoryN->Draw(opt.c_str());  	  	    
	}
	
      }

      //---------------------------------------------------------------
      if (hasVar[iHist] && uBands.size() > 1) {
	// Now go to second subdivision: relative uncertainties
	displays[iDisplay]->cd(iPlot+1);
	gPad->cd(2);	
	if (! vInitPad[iHist]) {
	  vAxisU[iHist].Draw();
	  vector<TGraph>& gVar = vGrTheoryVar[iHist];
	  for (unsigned int iBand=0; iBand< uBands.size(); iBand++) {
	    if (iBand == 0) {
	      gVar[iBand].SetLineColor(12);
	      gVar[iBand].SetLineWidth(3);
	    } else if (iBand == 1) {
	      gVar[iBand].SetLineColor(92);
	      gVar[iBand].SetFillColor(41);
	      gVar[iBand].SetFillStyle(3005);
	      gVar[iBand].SetLineWidth(3);
	    } else if (iBand == 2) {
	      gVar[iBand].SetLineColor(94);
	      gVar[iBand].SetLineWidth(2);
	    } else if (iBand == 3) {
	      gVar[iBand].SetLineColor(98);
	      gVar[iBand].SetLineStyle(3);
	      gVar[iBand].SetLineWidth(1);
	    } else if (iBand == 4) {
	      gVar[iBand].SetLineColor(96);
	      gVar[iBand].SetLineStyle(1);
	      gVar[iBand].SetLineWidth(1);
	    } else if (iBand == 5) {
	      gVar[iBand].SetLineColor(6);
	      gVar[iBand].SetLineStyle(2);
	      gVar[iBand].SetLineWidth(1);
	    }	   
	    gVar[iBand].Draw("sameline");
	  }
	  // Now write legends (so they appear on top of curves)
	  double xLeg = 0.15;
	  double xLegEnd = 0.20;
	  double yLeg = 0.07;
	  double yLegTop = 0.3;
	  for (unsigned int iBand=0; iBand< uBands.size(); iBand++) {
	    xLeg = xLegEnd - 0.025;
	    string name = uBands[iBand].name;
	    int cLength = name.size();
	    double length = cLength;
	    // Some characters fill less than others
	    for (int iPos=0; iPos < cLength; iPos++) {
	      string cPos = name.substr(iPos,1);
	      if (cPos == "#") length -= 1;
	      else if (cPos == "_") length -= 1.5;
	      else if (cPos == "^") length -= 1.5;
	      else if (cPos == "{") length -= 1.0;
	      else if (cPos == "}") length -= 1.0;
	      else if (cPos == "i") length -= 0.6;
	      else if (cPos == "j") length -= 0.4;
	      else if (cPos == "t") length -= 0.4;
	      else if (cPos == "f") length -= 0.4;
	      else if (cPos == "M") length += 0.1;
	      else if (cPos == "Q") length += 0.1;
	      else if (cPos == "O") length += 0.1;
	    }
	    // If dashed, make legend long enough for a few dashes
	    if (gVar[iBand].GetLineStyle() == 2) length +=2;
	    length = min(7.0,length);
	    length = max(4.2,length);
	    xLegEnd += 0.17*length/6.0;
	    // Add legend
	    TLegend *legU = new TLegend(xLeg,yLeg,xLegEnd,yLegTop,NULL,"rNDC");
	    legU->SetBorderSize(0);
	    legU->SetFillStyle(0);
	    string varLabel = uBands[iBand].name;
	    TLegendEntry *entry=legU->AddEntry("gr",varLabel.c_str(),"l");
	    entry->SetLineColor(gVar[iBand].GetLineColor());
	    entry->SetLineWidth(gVar[iBand].GetLineWidth());
	    entry->SetLineStyle(gVar[iBand].GetLineStyle());
	    entry->SetTextAlign(12);
	    entry->SetTextSize(0.23);
	    entry->SetTextFont(42);
	    entry->SetTextColor(1);
	    // Draw legend:
	    legU->Draw();	    
	  }
	} else {
	  gPad->Modified();
	}	
      }

      vInitPad[iHist] = true;

      // Return 
      return true;
    }

    //------------------------------------------------------------------------
    // addUncertainty
    // Store variation indices which together define an uncetainty band
    // and associate the band with a label. 
    void addUncertainty(string labelIn, int i1, int i2=-1, int i3=-1, int i4=-1) {
      vector<int> indxIn;
      if (i1 >= 0) indxIn.push_back(i1);
      if (i2 >= 0) indxIn.push_back(i2);
      if (i3 >= 0) indxIn.push_back(i3);
      if (i4 >= 0) indxIn.push_back(i4);
      addUncertainty(labelIn, indxIn);
    }

    void addUncertainty(string labelIn, vector<int> indicesIn ) {
      UncertaintyBand newBand(labelIn);
      newBand.indices = indicesIn;
      uBands.push_back(newBand);
    }

    //------------------------------------------------------------------------
    // Set MC Reference and/or Version names
    void setRefs(string refsIn) {
      mcRefs = refsIn;
    }

    // Set Label for theory curve(s)
    void setLabel(string labelIn) {
      mcLabel = labelIn;
    }

    // Set MC Reference and/or Version names
    void setOptStat(Int_t optStatIn) {
      optStat = optStatIn;
    }

    //------------------------------------------------------------------------
    // Set Color Scheme for all, or for an entire display, or for a specific plot
    void setColorScheme(int mCols, string displayNameIn="none", int iPlotIn=0) {

      // Define color schemes 
      Int_t thMarker, thStat, thUnc, thUncStat;
      if (mCols == 1) {
	// Black/Grey scheme 
	thMarker  =  12;
	thStat    =  14;
	thUnc     =  15;
	thUncStat =  17;
	refColor  =   2;
      } else if (mCols == 2) {
	// Red scheme
	thMarker  =  2;
	thStat    = 50;
	thUnc     = 42;
	thUncStat = 41;
	refColor  = 38;
      } else if (mCols == 3) {
	// Green scheme
	thMarker  = 32;
	thStat    = 31;
	thUnc     = 30;
	thUncStat = 29;
	refColor  =  4;
      } else {
	// Blue scheme (default)
	thMarker  = 61;
	thStat    = 64;
	thUnc     = 66;
	thUncStat = 67;
	refColor  =  2;
      }

      // Save to defaults
      if (displayNameIn == "none") {
	userColor1 = thMarker;
	userColor2 = thStat;
	userColor3 = thUnc;
	userColor4 = thUncStat;
      }

      // Check if only a specific display is desired modified?
      int iDisplay = -1;
      if (displayNameIn != "none") {
	for (unsigned int i=0; i<displayName.size(); i++) {
	  if (displayName[i] == displayNameIn) iDisplay = i;
	}
      }

      // Loop over histograms
      int iPlot = 0;
      for (unsigned int iHist=0; iHist < theoryMarkerColor.size(); iHist++) {
	// Restrict to specific display?
	if (iDisplay >= 0) {
	  if (vDisplayIndex[iHist] != iDisplay) continue;
	  // Restrict to specific plot on specific display?
	  iPlot++;
	  if (iPlotIn >= 1 && iPlotIn != iPlot) continue;
	}
	theoryMarkerColor[iHist]  = thMarker;
	theoryStatColor[iHist]    = thStat;
	theoryUncColor[iHist]     = thUnc;
	theoryUncStatColor[iHist] = thUncStat;
      }
      
    }

    //------------------------------------------------------------------------
    bool saveHistogram(TH1F* hist,string baseName,string varName,string ext) {

      // Construct name of output file and open it
      string fileName = baseName + "-" + varName + "." + ext;
      FILE * pFile = fopen (fileName.c_str(), "w");
      if (pFile == NULL) {
	perror ("Error opening file");
	return false;
      }
      else {
	// Number of bins to print
	int nBins = hist->GetNbinsX();
	for (int iBin=1; iBin<=nBins; iBin++){
	  double xLo = hist->GetBinLowEdge(iBin);
	  double xHi = xLo + hist->GetBinWidth(iBin);
	  double xC  = xLo + (xHi-xLo)/2.;
	  double y   = hist->GetBinContent(iBin);
	  double ey  = hist->GetBinError(iBin);
	  fprintf(pFile,"%14.6g %14.6g %14.6g %14.6g %14.6g %14.6g \n",xLo,xC,xHi,y,ey,ey); 
	}
	fprintf(pFile,"#          xLo             xC            xHi              y            +dy            -dy\n");
	fprintf(pFile,"SRC = %s\n",mcLabel.c_str());
	fprintf(pFile,"REF = %s\n",mcRefs.c_str());
	fprintf(pFile,"VAR = %s\n",varName.c_str());
	fprintf(pFile,"# Printed by VinciaRoot::saveHistogram()\n");	
	fclose(pFile);
	return true;
      }
      
    }

    //------------------------------------------------------------------------
    // Save current display
    bool saveDisplay(string displayNameIn, string firstName, string fileType="gif",
		     string option="Landscape") {
      for (int iDisplay=0; iDisplay<int(displayName.size()); iDisplay++) {
	if (displayName[iDisplay] == displayNameIn) {
	  string fName = firstName+"-"+displayNameIn;	  
	  // First loop for specific file types
	  if (fileType.find("eps") != string::npos) fName+=".eps";
	  else if (fileType.find("ps") != string::npos) fName+=".ps";
	  else if (fileType.find("gif") != string::npos) fName+=".gif";
	  else if (fileType.find("jpg") != string::npos) fName+=".jpg";
	  else if (fileType.find("png") != string::npos) fName+=".png";
	  else if (fileType.find("pdf") != string::npos) fName+=".pdf";
	  // Then try generic extension. Don't add dot if ext already has one.
	  else if (fileType.find(".") != string::npos) fName+=fileType;
	  else fName+="."+fileType;
	  displays[iDisplay]->SaveAs(fName.c_str(),option.c_str());
	  return true;
	}
      }
      return false;
    }

    //------------------------------------------------------------------------
    // Set verbose level
    void setVerbose(int verboseIn=1) {
      verbose=verboseIn;
    }

  private:
    
    vector<string> displayName;
    vector<TCanvas*> displays;
    vector<Int_t> theoryMarkerColor, theoryStatColor, theoryUncColor, 
      theoryUncStatColor;
    bool useDisplay, appIsInit;
    vector<int> xSiz, ySiz, xOffset, yOffset;
    vector<unsigned int> nColsSav, nRowsSav;
    // Vector of Root histrograms to include
    vector<int> vDisplayIndex;
    vector<TH1F*> vTheory;    
    map<int,vector<TH1F*> > vTheoryVar;
    map<int,vector<TGraph> > vGrTheoryVar;
    map<int,vector<TH1F*> > vThRef;
    // Local copies, for displays
    vector<TH1F> vAxis, vThCopyVar, vThCopyVarUnc, vThCopyLine;
    map<int,TH1F> vAxisN, vAxisU;
    vector<bool> vNorm, vStat, vLogY, vInitPad, hasVar;    
    vector<string> vData, vRef;
    vector<TLatex*> vTit;
    vector<double> vTheoryScale;
    // Separately indexed data sets, with vGData->index map
    map<int,TGraphAsymmErrors> vGrDataStat, vGrDataTot;
    map<int,TGraphAsymmErrors> vGrDataStatN, vGrDataTotN;
    map<int,TGraphAsymmErrors> vGrRefStat;
    map<int,TGraphAsymmErrors*> vGrTheoryN, vGrTheoryVarN, vGrRefStatN;
    map<int,string> vDataSrc, vDataRef, vRefSrc, vDataEcm;

    // Lines for dividing thrust
    TLine lineThrust34, lineThrust45;
    TLine lineMH34, lineMH45, lineC34;

    // Labels, etc
    string mcRefs, mcLabel;

    // Uncertainty band indices and names
    vector<UncertaintyBand> uBands;

    // Stat options
    Int_t optStat;

    // Colors
    Int_t dataColor, dataStatColor, dataSysColor;
    Int_t userColor1, userColor2, userColor3, userColor4;
    Int_t refColor;

    // Verbose level
    int verbose;
  
  };
  
}

#endif
