// VinciaAntenna.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2010 W.T. Giele, D.A. Kosower, P.Z. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Include the VinciaAntenna header file
#include "VinciaAntenna.h"

using namespace Pythia8;

namespace Vincia {
  
  //**********************************************************************
  
  // The Antenna class and its offspring
  
bool Antenna::init(bool checkAntenna) {

    // Check that pointers initialized
    if (! isInitPtr) {
      printOut("Antenna::init","Cannot initialize, pointers not set.");
      return false;
    }

    // Innocent until proven guilty
    isInit=true;

    // Force inputMode = 1 (double Laurent rep) for the time being
    inputMode = 1;

    // Set verbose level
    setParm(verbose,"verbose");

    // Read generic in- and out- states
    string o1,o2,n1,n2,n3,n4;
    if (name.find("Vincia:qQ:emit") != string::npos ) {
      o1="q";o2="qbar";n1="q";n2="g";n3="qbar";
    }
    else if (name.find("Vincia:qg:emit") != string::npos ) {
      o1="q";o2="g";n1="q";n2="g";n3="g";
    }
    else if (name.find("Vincia:gg:emit") != string::npos ) {
      o1="g";o2="g";n1="g";n2="g";n3="g";}      
    else if (name.find("Vincia:qg:split") != string::npos ) {
      o1="q";o2="g";n1="q";n2="qbar'";n3="q'";
    }
    else if (name.find("Vincia:gg:split") != string::npos ) {
      o1="g";o2="g";n1="g";n2="qbar";n3="q";
    }
    else {
      if (verbose >= 1)
	printErr(name+":init","unknown generic antenna");
      return false;
    }
    nameOld.push_back(o1);
    nameOld.push_back(o2);
    nameNew.push_back(n1);
    nameNew.push_back(n2);
    nameNew.push_back(n3);
    if (n4 != "") nameNew.push_back(n4);
    for (int ic=0;ic<int(nameOld.size());ic++) {
      if (nameOld[ic]=="q") idOld.push_back(1);
      if (nameOld[ic]=="qbar") idOld.push_back(-1);
      if (nameOld[ic]=="g") idOld.push_back(21);
      if (nameOld[ic]=="gamma") idOld.push_back(22);
    }
    for (int ic=0;ic<int(nameNew.size());ic++) {
      if (nameNew[ic]=="q") idNew.push_back(1);
      if (nameNew[ic]=="qbar") idNew.push_back(-1);
      if (nameNew[ic]=="g") idNew.push_back(21);
      if (nameNew[ic]=="gamma") idNew.push_back(22);
      if (nameNew[ic]=="q'") idNew.push_back(2);
      if (nameNew[ic]=="qbar'") idNew.push_back(-2);
    }
    if (humanName == name) humanName=name.substr(7,name.length());

    // First check whether antenna is 2->3
    if (nameOld.size() != 2 || nameNew.size() != 3) {
      if (verbose >= 1)
	printErr(name+":init","unknown antenna type: "+
		 num2str(int(idOld.size()))+"->"+num2str(int(idNew.size())));
      return false;
    }

    // Check if switched off
    if ( ! settingsPtr->flag(name)) {
        clear();
	isOnSav = false;
	setC(0,0,0.0);
	return true;
    } else {
      isOnSav = true;
    }

    // Set whether collinear and finite terms to be used in evolution
    isInit=isInit&&setParm(useCollinearTerms,"useCollinearTerms");
    isInit=isInit&&setParm(useFiniteTerms,"useFiniteTerms");
    // MR 2.3.2011: Set whether massive terms are used
    isInit=isInit&&setParm(useMassTerms,"useMassTerms");
    // Initialize parameter to use when checking for dead zones
    isInit=isInit&&setParm(deadZoneAvoidance,"deadZoneAvoidance");
    
    // Check charge conservation
    double sumQ=0.0;
    for (unsigned int ic=0;ic<idOld.size();ic++) 
      sumQ+=particleDataPtr->charge(idOld[ic]);
    for (unsigned int ic=0;ic<idNew.size();ic++) 
      sumQ-=particleDataPtr->charge(idNew[ic]);
    if (abs(sumQ) > TINY) {
      if (verbose >= 1)
	printErr(name+":init","antenna does not conserve charge");
      cout<<"   sumQ = "<<sumQ<<" Antenna ID :"<<idOld[1]<<" "
	  <<idOld[2]<<" -> "
	  <<idNew[1]<<" "<<idNew[2]<<" "<<idNew[3]<<endl;
      return false; 
    }

    // Read in antenna function terms (zero if switched off)    
    readC();  

    // Check that antenna function is positive definite 
    // and free of dead zones, if not rescale the finite terms
    if (checkAntenna) {
      bool initCheckPassed = checkInit(gettestMasses());
      if (!initCheckPassed) {
        return false;
      }
    }

    // Initialize evolution settings
    isInit=isInit&&setParm(evolutionType,"evolutionType"); 
    isInit=isInit&&setParm(kineMapType,"kineMapType");
    
    // Set chargeFactor (i.e., CA, CF, TR, etc.)
    isInit=isInit&&setParm(chargeFactor,"chargeFactor");
    
    // IR hadronization cutoff type and scale
    isInit=isInit&&setParm(cutoffType,"cutoffType");
    if (cutoffType == 0) cutoffType=evolutionType;
    isInit=isInit&&setParm(cutoffScale,"cutoffScale");
    
    // Return true or false, depending on how checks went.
    return isInit;
  }
  
  // the check for positivity and absence of dead zones called 
  // in init()
  bool Antenna::checkInit(vector< vector<double> > testMasses) {
    bool changedForPositivity = false;
    bool changedForNoDead = false;
    int nPointsCheck = 1000;
    
    for (unsigned int i = 0; i < testMasses.size(); ++i) {
      double mu1 = testMasses[i][0];
      double mu2 = testMasses[i][1];
      double mu3 = testMasses[i][2];
      if (verbose >= 4) {
        printOut(name + ":init", 
                 "checking with reduced masses " +
                 num2str( mu1 ) + ", " +
                 num2str( mu2 ) + ", " +
                 num2str( mu3 ) + ".");
      }
      // if all the positivity tests passed so far, call
      // IsPositive without instructing it to repair
      if (not changedForPositivity) {
        changedForPositivity = not IsPositive(nPointsCheck, 
                                              mu1, mu2, mu3, false);
      }
      // if a positivity test failed up to here, call IsPositive
      // instructing it to change the parameters if needed
      if (changedForPositivity) {
        if (not IsPositive(nPointsCheck,mu1,mu2,mu3,true)) {
          return false;
        };
      }
      // the analogue for dead zones
      if (not changedForNoDead) {
        changedForNoDead = not IsFreeOfDeadZones(
                                                 nPointsCheck, 
                                                 mu1, mu2, mu3, false);
      }
      if (changedForNoDead) {
        if (not IsFreeOfDeadZones(nPointsCheck, 
                                  mu1, mu2, mu3, true) ) {
          return false;
        };
      }
      
    }
    
    // give a message if coefficients were changed and the 
    // verbosity is high enough
    
    if (changedForPositivity) {
      if (verbose >= 1)
        printErr(humanName+":init","warning: Finite terms changed to ensure positivity");
      listCoefficients();
    } else if (changedForNoDead) {
      if (verbose >= 1)
        printErr(humanName+":init","warning: Finite terms changed to ensure absence of dead zones");
      listCoefficients();
    } else {
      if (verbose >= 4)
        printOut(humanName+":init","ok: positive definite, no dead zones");
    }
    // if the function arrives here, things are OK
    return true;
  }
  
  bool Antenna::setParm(bool& par, string parName) {
  bool found=true;
  if (settingsPtr->isFlag(name+":"+parName)) 
    par = settingsPtr->flag(name+":"+parName);
  else if (settingsPtr->isFlag("VinciaShower:"+parName)) 
    par = settingsPtr->flag("VinciaShower:"+parName);
  else if (settingsPtr->isFlag("Vincia:"+parName)) 
    par = settingsPtr->flag("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName); }
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found"); }
  return found;
}

bool Antenna::setParm(int& par, string parName) {
  bool found=true;
  if (settingsPtr->isMode(name+":"+parName)) 
    par = settingsPtr->mode(name+":"+parName);
  else if (settingsPtr->isMode("VinciaShower:"+parName)) 
    par = settingsPtr->mode("VinciaShower:"+parName);
  else if (settingsPtr->isMode("Vincia:"+parName)) 
    par = settingsPtr->mode("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose>= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(double& par, string parName) {
  bool found=true;
  if (settingsPtr->isParm(name+":"+parName)) 
    par = settingsPtr->parm(name+":"+parName);
  else if (settingsPtr->isParm("VinciaShower:"+parName)) 
    par = settingsPtr->parm("VinciaShower:"+parName);
  else if (settingsPtr->isParm("Vincia:"+parName)) 
    par = settingsPtr->parm("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

bool Antenna::setParm(string& par, string parName) {
  bool found=true;
  if (settingsPtr->isWord(name+":"+parName)) 
    par = settingsPtr->word(name+":"+parName);
  else if (settingsPtr->isWord("VinciaShower:"+parName)) 
    par = settingsPtr->word("VinciaShower:"+parName);
  else if (settingsPtr->isWord("Vincia:"+parName)) 
    par = settingsPtr->word("Vincia:"+parName);
  else found=false;
  if (found) {if (verbose >= 4) printOut(humanName+":setParm","ok: "+parName);}
  else {if (verbose >= 1) printErr(humanName+":setParm",parName+": not found");}
  return found;
}

// Read in massless Laurent coefficients from Settings database

void Antenna::readC() {
  if (verbose >= 5)
    printOut(humanName+":readC","reading antenna coefficients ");
  // Antenna function
  clear();
 
  // read useFiniteTerms as well
  if ( settingsPtr->isFlag("VinciaShower:UseFiniteTerms") ) {
    useFiniteTerms = settingsPtr->flag("VinciaShower:UseFiniteTerms");
  }
 
  // Input as Laurent Series
  if (inputMode == 1) {

    //------------------------------------------------------------------
    // 1st part: read in massless antenna coefficients

    for (int ic=-1;ic<=LaurentIndexMax;ic++) {
      string baseName("LaurentC("+num2str(ic,0)+",");
      for (int jc=-1;jc<=LaurentIndexMax;jc++) {
	string CName(baseName+num2str(jc,0)+")"); 
	if (settingsPtr->isParm(name+":"+CName)) {	
	  if (setC(ic,jc,settingsPtr->parm(name+":"+CName))) {
	    int verb = (getC(ic,jc) == 0.0) ? 9 : 6; 
	    if (verbose >= verb)
	      printOut(humanName+":readC",
		     CName+" = "+num2str(getC(ic,jc),6));
	  }
	} 
      } 
    }
  } else {
    // Enter alternative input forms here
  }
}

bool Antenna::setC(int ic,int jc, double cIn) {
  if (inputMode == 1) {
    // Input as Laurent series
    if (ic < -1 or jc < -1) {
      if (verbose >= 1)
	printErr(humanName+":setC",
		 "leading term is (-1,-1) : ignoring entry ("
		 +num2str(ic)+","+num2str(jc)+")");
      return false;
    }
    else if (cIn == 0) {
      if (LaurentC.find(ic) != LaurentC.end()) {
	if (LaurentC[ic].find(jc) != LaurentC[ic].end()) LaurentC[ic].erase(jc);
      } 
    }
    else if (ic == -1 && jc == -1 && cIn>0) LaurentC[ic][jc]=cIn;
    else if (jc == -1 || ic == -1) LaurentC[ic][jc]=cIn;
    else LaurentC[ic][jc]=cIn;
  } else {
    // Enter alternative input modes here
    if (verbose >= 1)
      printErr(humanName+":setC",
	     "sorry, desired inputMode not recognized.");
    return false;
  }
  return true;
}

double Antenna::getC(int ic,int jc) {     
  if (inputMode == 1) {
    // Input as Laurent series
    map<int,map<int,double> >::iterator it;
    map<int,double>::iterator jt;
    it=LaurentC.find(ic);
    if (it==LaurentC.end()) return 0.0;
    jt=LaurentC[ic].find(jc);
    if (jt==LaurentC[ic].end()) return 0.0;
    return LaurentC[ic][jc];
  } else {
    // Enter alternative input modes here
    return 0.0;
  }
}
  
void Antenna::clear() {
  if (inputMode == 1) {
    // Clear Laurent series coeffiecents
    LaurentC.clear(); 
  } else {
    // Enter alternative input modes here
  }
}

  double Antenna::a2(
    double y12, double y23, double massmu1, double massmu2,
    double massmu3, bool forcepositive
    ) {
    // First check if switched off
    if ( !isOn() ) return 0.0;
    
    // calculate the missing invariant
    double y13 =
    1 - y12 - y23 - pow2( massmu1 ) - pow2( massmu2 ) - pow2( massmu3 );
    // complain if y13 is below a value approaching machine precision
    if (y13 - 2*massmu1*massmu3 < -TINY) {
      if (verbose>=3) {
        printOut("VinciaAntenna::a2", "y13 < 2*mu1*mu3 occurred");
        cout << "   y12 = " << y12 ;
        cout << " y23 = " << y23 ;
        cout << " y13 = " << y13 << "\n";
        cout << "   mu1 = " << massmu1 ;
        cout << " mu2 = " << massmu2 ;
        cout << " mu3 = " << massmu3 <<endl;      
      }
      return 0.0;
    } 
    // Tolerate values slightly below y13min (due to rounding), 
    // but set them = 2*massmu1*massmu3
    else if (y13 < 2*massmu1*massmu3) {
      y13 = 2*massmu1*massmu3;
    }
    
    // MR 21.12.2010: Use Gram determinant to check whether we are too
    // close to the boundary of the phase space.
    // Check for gDet / Ecm^6 < TINY^3 because the Gram determinant has 
    // mass dimension 6
    // have to check y12min <= y12 <= y12max first because
    // for y12 > y12max, a region with positive gram
    // determinant opens up
    double y12min = 2*massmu1*massmu2;
    double y12max = pow2( 1-massmu3 ) - pow2(massmu1) - pow2(massmu2);
    if ( (y12 - y12min < 0.5*TINY ) || (y12max - y12 < 0.5*TINY) ) {
      if (verbose >= 3) {
        printOut(name+":a2",
                 "y12 = " + num2str(y12, 4) + " outside physical phase "
                 + "space or too close to boundaries y12min = " 
                 + num2str(y12min, 4) + ", y12max = " 
                 + num2str(y12max, 4) + "}, return 0.");
      }
      return 0.0;
    }
    double gdetdimless = gDet(y12, y23, y13, massmu1, massmu2, massmu3);
    if (gdetdimless< pow3(TINY)) {
      if (verbose >= 3) {
        printOut(name+":a2", 
                 "Gram determinant " + num2str(gdetdimless, 15) 
                 + " < " + num2str(pow3(TINY), 15)  + ", return 0\n"
                 + "\ty12 = " + num2str(y12, 15) + ",\n"
                 + "\ty23 = " + num2str(y23, 15) + ",\n"
                 + "\ty13 = " + num2str(y13, 15) + ",\n"
                 + "\tmu1 = " + num2str(massmu1, 15) + ",\n"
                 + "\tmu2 = " + num2str(massmu2, 15) + ",\n"
                 + "\tmu3 = " + num2str(massmu3, 15)
                 );
      }
      return 0.0;
    }
    
    double val = a2Raw(y12,y13,y23,massmu1,massmu2,massmu3);
    if (forcepositive) {
      return max(val, 0.0);
    }
    else {
      return val;
    }
  }
  
  // Antenna does only take into account the C, D, E, K coefficients
  double Antenna::a2Raw(
    double y12, double, double y23, 
    double, double, double
    ){
    // initialize the return value
    double val=0.0;
    if (inputMode == 1) {
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      // Add C(i,j) terms
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
        int ic=it->first;
        for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
          int jc=jt->first;
	  bool useTerm = false;
	  // Always add Eikonal term
	  if (ic == -1 && jc == -1) useTerm = true;
	  // Always add next-to-Eikonal terms
	  else if (min(ic,jc) == -1 && max(ic,jc) == 0) useTerm = true;
	  // Only add collinear terms if they are "on"
	  else if ( (ic < 0 || jc < 0) && useCollinearTerms) useTerm = true;
          // Only add finite terms if they are "on"
          else if ( ic >= 0 && jc >= 0 && useFiniteTerms) useTerm = true;
	  if (useTerm) { 
	    double c=jt->second;      
	    val += c * pow(y12,ic) * pow(y23,jc);
	  }
        }      
      }
    } // closing of "if InputMode == 1 ..."
    
    // if InputMode is not 1
    else {
      // Enter alternative input forms here
    }
    return val;
  }
  
  // MR 21.12.2010: Changed the check function to rely on IsPositive
  // and IsFreeOfDeadZones. This loses a factor of 2 in efficiency
  // (the same phase space point is checked twice, once by IsPositive,
  // once by IsFreeOfDeadZones), but it makes the function easier to read.

  bool Antenna::check(
    int nPointsCheck, double mu1, double mu2, double mu3, bool dorepair
    ) {
    bool isPositive = 
          this->IsPositive(nPointsCheck, mu1, mu2, mu3, dorepair);
    bool isFreeOfDeadZones = 
          this->IsFreeOfDeadZones(nPointsCheck, mu2, mu2, mu3, dorepair);
    return (isPositive && isFreeOfDeadZones);
  }
  
// MR 21.12.2010: IsPositive checks whether an antenna is positive
// given a set of masses. It has an optional argument dorepair (default 
// false) which governs whether the finite coefficients are changed 
// to make the antenna positive.
bool Antenna::IsPositive( 
  int nPointsCheck, double mu1, double mu2, double mu3, bool dorepair
  ) {
  // initialize the flags
  bool isPos = true;
  
  // don't want the antenna to complain, set verbosity to -1
  int verboseSav = verbose;
  verbose = -1;
  // calculate the phase space boundaries
  double y12min = 2*mu1*mu2;
  double y12max = pow2(1 - mu3) - pow2(mu1) - pow2(mu2);
  // loop through the phase space
  int numberofsteps = floor( sqrt( nPointsCheck ) );
  double y12StepWidth = (y12max - y12min)/numberofsteps;
  // i = 0 and i <= numberofsteps are on the boundary of the phase
  // space, omit them
  for (int i=1; i<=numberofsteps-1; i++) {
    double y12 = y12min + i*y12StepWidth;
    double y23min = y23tildeMin(y12, mu1, mu2, mu3);
    double y23max = y23tildeMax(y12, mu1, mu2, mu3);
    double y23StepWidth = (y23max - y23min)/numberofsteps;
    // j=0 and j=numberofsteps are on the boundary of the phase space,
    // omit them
    for (int jc=1; jc<=numberofsteps-1; jc++) {
      double y23 = y23min + jc*y23StepWidth;
      // Evaluate a2 with and without finite terms, enable
      // a2 to return negative values
      double aFull   = a2(y12,y23, mu1, mu2, mu3, false);
      bool useFinSav = useFiniteTerms;
      useFiniteTerms = false;
      double aSing   = a2(y12,y23, mu1, mu2, mu3, false);
      // reset the use of the finite terms
      useFiniteTerms = useFinSav;
      if ( aFull <= - TINY ) {
	isPos = false;
	if (dorepair) {
	  // if the singular part of the antenna is negative 
	  // and the function is supposed to repair the 
	  // antenna, there is a problem
	  if (aSing <= -TINY) {
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 1) {
	      printOut(name+"a2:IsPositive",
		       string("Singular part of antenna ")
		       + " negative.\n"
		       + "\ty12 = " + num2str(y12, 4) +"\n"
		       + "\ty23 = " + num2str(y12, 4) +"\n"
		       + "\tmu1 = " + num2str(mu1, 4) +"\n"
		       + "\tmu2 = " + num2str(mu2, 4) +"\n"
		       + "\tmu3 = " + num2str(mu3, 4) +"\n"
		       + "\taFull = " + num2str(aFull, 4) +"\n"
		       + "\taSing = " + num2str(aSing, 4) +"\n"
		       );
	    }
	    verbose = verboseSav;
	    return false;
	  }
	  else {
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 5) {
	      printOut(name+":IsPositive",
		       "rescaling finite terms to make "
		       + string("antenna positive")
		       );
	    }
	    bool rescaleSucessful =
	      this->rescaleFiniteTerms(0.9);
	    if (rescaleSucessful) {
	      // restart the check with the changed 
	      // finite terms
	      if (verboseSav >= 3) {
		printOut(name+":IsPositive",
			 "new coefficients are:");
		this->listCoefficients();
	      }
	      verbose = verboseSav;
	      return this->IsPositive(nPointsCheck, mu1, 
				      mu2, mu3, true);
	    }
	    else {
	      // if it has not been possible to rescale
	      // the finite terms, complain and return
	      // false
	      // verbose has been set to -1, 
	      // use verboseSav
	      if (verboseSav >= 1) {
		printOut(name+"a2:IsPositive",
			 "Could not rescale finite "
			 + string("terms.\n") );
	      }
	      verbose = verboseSav;
	      return false;
	    }
	  }
	}
	else {
	  // verbose has been set to -1, use verboseSav
	  if (verboseSav >= 3) {
	    cout << "IsPositive::negative value" << endl;
	    cout << "Antenna name = " << getHumanName();
	    cout << endl;
	    cout << scientific;
	    cout << setprecision(20);
	    cout << "mu1 = " << mu1 << endl;
	    cout << "mu2 = " << mu2 << endl;
	    cout << "mu3 = " << mu3 << endl;
	    cout << "y12 = " << y12 << endl;
	    cout << "y23 = " << y23 << endl;
	    cout << "full antenna = " << aFull << endl;
	    cout << "singular part = " << aSing << endl;
	  }
	}
      }
    } // end of for (int j=0; j<=numberofsteps...
  }
  // reset the verbosity
  verbose = verboseSav;
  return isPos;
}

// MR 21.12.2010: IsFreeOfDeadZones checks whether an antenna has dead
// zones given a set of masses. It has an optional argument dorepair 
// (default false) which governs whether the finite coefficients are 
// changed to make the dead zones go away.
bool Antenna::IsFreeOfDeadZones( int nPointsCheck, double mu1, 
				 double mu2, double mu3, 
				 bool dorepair ) {
  // initialize the flags
  bool isFreeOfDeadZones = true;
  
  // don't want the antenna to complain, set verbosity to -1
  int verboseSav = verbose;
  verbose = -1;
  
  // calculate the phase space boundaries
  double y12min = 2*mu1*mu2;
  double y12max = pow2(1 - mu3) - pow2(mu1) - pow2(mu2);
  // loop through the phase space
  int numberofsteps = floor( sqrt( nPointsCheck ) );
  double y12StepWidth = (y12max - y12min)/numberofsteps;
  // i=0 and i=numberofsteps are the boundary of the phase space,
  // omit them
  for (int i=1; i<=numberofsteps-1; i++) {
    double y12 = y12min + i*y12StepWidth;
    // make sure y12 is positive
    y12 = max(y12,0.);
    double y23min = y23tildeMin(y12, mu1, mu2, mu3);
    double y23max = y23tildeMax(y12, mu1, mu2, mu3);
    double y23StepWidth = (y23max - y23min)/numberofsteps;
    // j=0 and j=numberofsteps are the boundary of the phase space,
    // omit them
    for (int jc=1; jc<=numberofsteps-1; jc++) {
      double y23 = y23min + jc*y23StepWidth;
      // make sure y23 is positive
      y23 = max(y23, 0.);
      // Evaluate a2 with and without finite terms, enable
      // a2 to return negative values
      double aFull   = a2(y12,y23, mu1, mu2, mu3, false);
      bool useFinSav = useFiniteTerms;
      useFiniteTerms = false;
      double aSing   = a2(y12,y23, mu1, mu2, mu3, false);
      // reset the use of the finite terms
      useFiniteTerms = useFinSav;
      // only do something if aFull< deadZoneAvoidance*aSing
      if ( aFull < deadZoneAvoidance*aSing ) {
	isFreeOfDeadZones = false;
	if (dorepair) {
	  // rescale
	  if ( this->rescaleFiniteTerms(0.9) ) {
	    // restart the check with the changed 
	    // finite terms
	    verbose = verboseSav;
	    return this->IsFreeOfDeadZones(nPointsCheck, 
					   mu1, mu2, mu3, true);
	  }
	  else {
	    // if it has not been possible to rescale
	    // the finite terms, complain and return
	    // false
	    // verbose has been set to -1, use verboseSav
	    if (verboseSav >= 1) {
	      printOut(name+"a2:IsFreeOfDeadZones",
		       "Could not rescale finite terms.\n");
	    }
	    verbose = verboseSav;
	    return false;
	  }
	}
	else {
	  // verbose has been set to -1, use verboseSav
	  if (verboseSav >= 3) {
	    cout << "IsFreeOfDeadZones::dead zone" << endl;
	    cout << "Antenna name = " << getHumanName();
	    cout << endl;
	    cout << scientific;
	    cout << setprecision(20);
	    cout << "mu1 = " << mu1 << endl;
	    cout << "mu2 = " << mu2 << endl;
	    cout << "mu3 = " << mu3 << endl;
	    cout << "y12 = " << y12 << endl;
	    cout << "y23 = " << y23 << endl;
	    cout << "full antenna = " << aFull << endl;
	    cout << "singular part = " << aSing << endl;
	    cout << "dead Zone avoidance factor = " 
		 << deadZoneAvoidance << endl;
	  }
	}
      }
    }
  }
  // reset the verbosity
  verbose = verboseSav;
  return isFreeOfDeadZones;
}

  bool Antenna::rescaleFiniteTerms(double rescalingFactor) {
    if (inputMode == 1) {
      // list the old parameter values
      if (verbose >= 5) {
	printOut(name+":rescaleFiniteTerms", 
		 string("reducing magnitude of finite terms.\n")
		 + "old parameter values:\n"
		 );
	listCoefficients();
      }
      // scale C(i,j) terms
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
	int ic=it->first;
	for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
	  int jc=jt->first;  
	  if ( ic >= 0 && jc >= 0) {
	    double oldcoeff = jt->second;
	    double newabs = rescalingFactor*abs(oldcoeff);
	    // round the new absolute value down to the closest
	    // multiple of abs(1-rescalingFactor)
	    double minchange = abs(1-rescalingFactor);
	    newabs = floor( newabs/minchange )*minchange;
	    setC(ic, jc,oldcoeff*newabs/abs(oldcoeff) );
	  }
	}
      } // end of loop through C(i,j) terms
      return true;
    }
    else {
      if (verbose >= 3) {
	printOut(name+":rescaleFiniteTerms", 
		 "inputMode " + num2str( inputMode )
		 + " not recognized.");
      }
      return false;
    }
  }

// MR 16.7.2010: for debugging purposes, have the antenna list
// its coefficients on an output stream
void Antenna::listCoefficients(ostream& os) {
  os << setprecision(3);
  os << fixed;
  os << "antenna " << getHumanName() << "\n";
  
  os << "use finite terms: ";
  if (useFiniteTerms) {
    os << "True" << endl;
  }
  else {
    os << "False" << endl;
  }
  
  map<int,map<int,double> >::iterator it;
  map<int,double>::iterator jt;
  // C(i,j) terms, first the fixed ones, then the finite ones
  for (it=LaurentC.begin();it!=LaurentC.end();++it) {
    int i=it->first;
    for (jt=LaurentC[i].begin();jt!=LaurentC[i].end();++jt) {
      int jc=jt->first;  
      double c=jt->second;
      if (i < 0 || jc < 0) {
        os << "LaurentC(" << i << "," << jc << ") = " << c << "\n";
      }
    }      
  } 
  for (it=LaurentC.begin();it!=LaurentC.end();++it) {
    int i=it->first;
    for (jt=LaurentC[i].begin();jt!=LaurentC[i].end();++jt) {
      int jc=jt->first;  
      double c=jt->second;
      if (i >= 0 && jc >= 0) {
        os << "LaurentC(" << i << "," << jc << ") = " << c << "\n";
      }
    }      
  }
}

// test masses for Antenna base are {0,0,0}
vector< vector<double> > Antenna::gettestMasses() {
  vector< vector<double> > retVec;
  retVec.push_back( vector<double>(3, 0.0) );
  return retVec;
}

    
  void AntennaqqEmit::readC() {
    Antenna::readC();
    
    // read also useMassTerms
    if ( settingsPtr->isFlag(name + ":UseMassTerms") ) {
      useMassTerms = settingsPtr->flag(name + ":UseMassTerms");
    }
    
    // use getParms
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu1ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int mu3ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int y12ord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int y23ord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setM(mu1ord,mu3ord,y12ord,y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(mu1ord, mu3ord, y12ord, y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+" = "+num2str(getM(mu1ord, mu3ord, y12ord, y23ord),6)
          );
      }
    }
    const string 
      denominatorMassPrefactorName("DenominatorMassPrefactor");
    if (settingsPtr->isParm(name + ":" + denominatorMassPrefactorName)){
      if (
        setDenominatorMassPrefactor(
          settingsPtr->parm(name+":"+denominatorMassPrefactorName))){
        int verb = (denominatorMassPrefactor == 0.0) ? 8 : 6;
        if (verbose >= verb) {
          printOut(humanName+":readC",
            denominatorMassPrefactorName + "=" 
            + num2str(getDenominatorMassPrefactor(), 6) );
        }
      }
    }
  }
  
  void AntennaqqEmit::clear(){
    if (inputMode == 1) {
      // Clear Laurent series coeffiecents
      Antenna::clear();
      LaurentM.clear();
      setDenominatorMassPrefactor(0.0);
    } else {
      // Enter alternative input modes here
    }
  }  
  
  void AntennaqqEmit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    os << "denominator mass prefactor: " << denominatorMassPrefactor
      << endl;
    for (map<FourInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      FourInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mIt->second << endl;
    }
  }
  
  bool AntennaqqEmit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    if (inputMode == 1) {
      for (map< FourInt, double>::iterator pt=LaurentM.begin();
	   pt!=LaurentM.end();++pt) {
        int mu1ord = pt->first.int1;
        int mu3ord = pt->first.int2;
        int y12ord = pt->first.int3;
        int y23ord = pt->first.int4;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setM(
        mu1ord, mu3ord, y12ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      return baseOK;
    }
    else {
      if (verbose >= 3) {
        printOut(name+":rescaleFiniteTerms", 
                 "inputMode " + num2str( inputMode )
                 + " not recognized.");
      }
      return false;
    }
  }
  
  double AntennaqqEmit::getM(
    int mu1ord, int mu3ord, int y12ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentM.find(FourInt(mu1ord, mu3ord, y12ord, y23ord));
    if (pos == LaurentM.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaqqEmit::setM(
    int mu1ord, int mu3ord, int y12ord, int y23ord, double mIn
    ){
    if (inputMode == 1) {
      FourInt index(mu1ord, mu3ord, y12ord, y23ord);
      if (mIn == 0.0) {
        LaurentM.erase(index);
      }
      else {
        LaurentM[index] = mIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setM",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  double AntennaqqEmit::getDenominatorMassPrefactor() {
    if (inputMode == 1) {
      return denominatorMassPrefactor;
    } else {
      // Enter alternative input modes here
      return 0.0;
    }
  }
  
   bool AntennaqqEmit::setDenominatorMassPrefactor(double pfIn) {
    if (inputMode == 1) {
      // Input as Laurent series
      denominatorMassPrefactor=pfIn;
    } else {
      // Enter alternative input modes here
      if (verbose >= 1) 
        printErr(humanName+":setDenominatorMassPrefactor",
                 "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  vector< vector<double> > AntennaqqEmit::gettestMasses() {
    // masses for a03 (m1,0,m2)
      const double a03_masses_input[][2] = 
      {	
        {0,0}, {0.01,0}, {0.05,0}, {0.3,0}, {0.5,0}, {0.9,0}, {0.95,0},
        {0.99,0}, {1,0},
        {0.01,0.01}, {0.05,0.05}, {0.3,0.3}, {0.45,0.45}, {0.49,0.49},
        {0.01,0.3}, {0.05,0.3}, {0.1,0.3},
        {0.01,0.95}, {0.01,0.99}, {0.05,0.9}, {0.05,0.95}, 
        {0.1,0.8}, {0.1,0.9}, {0.3,0.6}, {0.3,0.7}
      };
    vector< vector<double> > a03_masses;
    for (size_t i=0; i<sizeof(a03_masses_input)/(2*sizeof(double)); i++) {
      vector<double> prov_masses;
      prov_masses.push_back(a03_masses_input[i][0]);
      prov_masses.push_back(0);
      prov_masses.push_back(a03_masses_input[i][1]);
      a03_masses.push_back( prov_masses );
    }
    return a03_masses;
  }

  double AntennaqqEmit::a2Raw(
    double y12, double y13, double y23, double mu1, double, 
    double mu3
    ){
    // call the function with masses zero if useMassTerms is false
    if ( !useMassTerms && (mu1 != 0.0 || mu3 != 0.0) ) {
      return a2Raw(y12, y13, y23, 0.0, 0.0, 0.0);
    }
    double eikonal = 
      2*y13/(y12*y23) - 2*pow2(mu1)/pow2(y12) - 2*pow2(mu3)/pow2(y23);
    double denominator = y12+y13+y23+ denominatorMassPrefactor*mu1*mu3;
    double collinear = 0.0;
    if (useCollinearTerms) {
      collinear = y12/y23 + y23/y12;
    }
    double fin = 0.0;
    if (useFiniteTerms) {
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      // Add C(i,j) terms
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
        int ic=it->first;
        for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
          int jc=jt->first;  
          // Only add finite terms
          if (ic >= 0 && jc >= 0) {
            double c=jt->second;
            // average c with charge-conjugate coefficient
            if (ic != jc) {
              c = 0.5*( c + getC(jc,ic) );
            }      
            fin += c * pow(y12,ic) * pow(y23,jc);
          }
        }      
      }
      if (max(mu1, mu3) > 0.0) {
        for (map< FourInt, double>::iterator mIt=LaurentM.begin();
             mIt!=LaurentM.end();++mIt) {
          int mu1ord = mIt->first.int1;
          int mu3ord = mIt->first.int2;
          int y12ord = mIt->first.int3;
          int y23ord = mIt->first.int4;
          double presentcoeff = mIt->second;
          // ignore negative y12 or y23 or exponents silently, 
          // as well as nonpositive mu1ord and mu3ord 
          if (
            min(y12ord, y23ord) >= 0 && mu1ord >= 0 && mu3ord >= 0
              && max(mu1ord, mu3ord) >= 1
            ) {
            fin += 
              presentcoeff * pow(y12, y12ord)*pow(y23, y23ord) 
              * pow(mu1, mu1ord)* pow(mu3, mu3ord);
            // enforce charge conjugation symmetry by adding the 
            // mu1ord <-> mu3ord coefficient 
            // for non-symmetric coefficients
            if (mu1ord != mu3ord) {
              fin += presentcoeff*pow(y12, y12ord)*pow(y23, y23ord)
                      *pow(mu1, mu3ord)*pow(mu3, mu1ord);
            }
            // analogous for y12ord <-> y23ord
            if (y12ord != y23ord) {
              fin += presentcoeff*pow(y12, y23ord)*pow(y23, y12ord)
                      *pow(mu1, mu1ord)*pow(mu3, mu3ord);
            }
            // and finally the combined switch
            if (mu1ord != mu3ord && y12ord != y23ord) {
              fin += presentcoeff*pow(y12, y23ord)*pow(y23, y12ord)
                      *pow(mu1, mu3ord)*pow(mu3, mu1ord);
            }
          }
        }
      }
    } // end of if (useFiniteTerms)
    return eikonal + 1/denominator*(collinear + fin);
  }
  
  void AntennaqgEmit::readC() {
    Antenna::readC();
    
    // read in collinear gluon splitting parameter
    // this is a global parameter, it cannot have different values for
    // different antennae
    const string cgsname("Vincia:CollinearGluonSplitting");
    if ( settingsPtr->isParm(cgsname) ) {	
      if ( setCollinearGluonSplitting(settingsPtr->parm(cgsname)) ) {
	int verb = (getCollinearGluonSplitting() == 0.0) ? 8 : 6; 
	if (verbose >= verb)
	  printOut(
            humanName+":readC",
            cgsname+" = "+num2str(getCollinearGluonSplitting(),6)
            );
      }
    }
    
    // read also useMassTerms
    if ( settingsPtr->isFlag(name + ":UseMassTerms") ) {
      useMassTerms = settingsPtr->flag(name + ":UseMassTerms");
    }
    
    // use getParms
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int muord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int y12ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int y23ord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      setM(muord,y12ord,y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(muord, y12ord, y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+" = "+num2str(getM(muord, y12ord, y23ord),6)
          );
      }
    }
    
    // analogous for the mtilde coefficients
    map<string,Parm> mtildecoeffs = 
      settingsPtr->getParmMap(name+":LaurentMtilde(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mtildecoeffs.begin();
      pt!=mtildecoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int muord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int y12ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int y13ord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int y23ord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setMtilde(muord,y12ord,y13ord,y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getMtilde(muord,y12ord,y13ord,y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+" = "+num2str(getMtilde(muord,y12ord,y13ord,y23ord),6)
          );
      }
    }
  }
  
  void AntennaqgEmit::clear(){
    if (inputMode == 1) {
      // Clear Laurent series coeffiecents
      Antenna::clear();
      LaurentM.clear();
      LaurentMtilde.clear();
      setCollinearGluonSplitting(0.0);
    } else {
      // Enter alternative input modes here
    }
  }  
  
  void AntennaqgEmit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    os << "collinear gluon splitting: " << collinearGluonSplitting
      << endl;
    for (map<ThreeInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      ThreeInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << "): " 
        << mIt->second << endl;
    }
    for (map<FourInt, double>::iterator mtildeIt = LaurentMtilde.begin(); 
      mtildeIt != LaurentMtilde.end(); ++mtildeIt
      ) {
      FourInt indx = mtildeIt->first;
      os << "LaurentMtilde(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mtildeIt->second << endl;
    }
  }
  
  bool AntennaqgEmit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    if (inputMode == 1) {
      for (map< ThreeInt, double>::iterator pt=LaurentM.begin();
	   pt!=LaurentM.end();++pt) {
        int muord = pt->first.int1;
        int y12ord = pt->first.int2;
        int y23ord = pt->first.int3;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setM(muord,y12ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      for (map< FourInt, double>::iterator pt=LaurentMtilde.begin();
	   pt!=LaurentMtilde.end();++pt) {
        int muord = pt->first.int1;
        int y12ord = pt->first.int2;
        int y13ord = pt->first.int3;
        int y23ord = pt->first.int4;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setMtilde(
          muord,y12ord, y13ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      return baseOK;
    }
    else {
      if (verbose >= 3) {
        printOut(name+":rescaleFiniteTerms", 
                 "inputMode " + num2str( inputMode )
                 + " not recognized.");
      }
      return false;
    }
  }
  
  double AntennaqgEmit::getCollinearGluonSplitting() {
    return collinearGluonSplitting;
  }
  
  bool AntennaqgEmit::setCollinearGluonSplitting(double s) {
    if (inputMode == 1) {
      // check for positivity
      if (s >= 0.0) {
        collinearGluonSplitting = s;
        return true;
      }
      return false;
    }
    return false;
  }
  
  double AntennaqgEmit::getM(
    int muord, int y12ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map<ThreeInt, double>::iterator pos = 
      LaurentM.find(ThreeInt(muord, y12ord, y23ord));
    if (pos == LaurentM.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaqgEmit::setM(
    int muord, int y12ord, int y23ord, double mIn
    ){
    if (inputMode == 1) {
      ThreeInt index(muord, y12ord, y23ord);
      if (mIn == 0.0) {
        LaurentM.erase(index);
      }
      else {
        LaurentM[index] = mIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setM",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  double AntennaqgEmit::getMtilde(
    int muord, int y12ord, int y13ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentMtilde.find(FourInt(muord, y12ord, y13ord, y23ord));
    if (pos == LaurentMtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaqgEmit::setMtilde(
    int muord, int y12ord, int y13ord, int y23ord, double mIn
    ){
    if (inputMode == 1) {
      FourInt index(muord, y12ord, y13ord, y23ord);
      if (mIn == 0.0) {
        LaurentMtilde.erase(index);
      }
      else {
        LaurentMtilde[index] = mIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setMtilde",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  vector< vector<double> > AntennaqgEmit::gettestMasses() {
     // masses for d03 (m,0,0)
    const double d03_masses_input[] = 
      {0, 0.01, 0.05, 0.1, 0.3, 0.5, 0.7, 0.9, 0.95, 0.99, 1};
    vector< vector<double> > d03_masses;
    for (size_t i=0; i<sizeof(d03_masses_input)/sizeof(double); i++) {
      vector<double> prov_masses;
      prov_masses.push_back(d03_masses_input[i]);
      prov_masses.push_back(0);
      prov_masses.push_back(0);
      d03_masses.push_back( prov_masses );
    }
    return d03_masses;
  }
  
  double AntennaqgEmit::a2Raw(
    double y12, double y13, double y23, double mu1, double, double
    ) {
    // call the function with mass zero if useMassTerms is false
    if ( !useMassTerms && (mu1 != 0.0 ) ) {
      return a2Raw(y12, y13, y23, 0.0, 0.0, 0.0);
    }
    double eikonal = 
      2*y13/(y12*y23) - 2*pow2(mu1)/pow2(y12);
    double den = y12 + y13 + y23;
    double collinear = 0.0;
    if (useCollinearTerms) {
      collinear = 
      y23/y12 + y12/y23*y13/den + collinearGluonSplitting*(y12-y13)/y23;
    } 
    double fin = 0.0;
    if (useFiniteTerms) {
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      // Add C(i,j) terms
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
        int ic=it->first;
        for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
          int jc=jt->first;  
          // Only add finite terms
          if (ic >= 0 && jc >= 0) {
            double c=jt->second;      
            fin += c * pow(y12,ic) * pow(y23,jc);
          }
        }      
      }
      if (mu1 > 0.0) {
        for (map< ThreeInt, double>::iterator mIt=LaurentM.begin();
             mIt!=LaurentM.end();++mIt) {
          int muord = mIt->first.int1;
          int y12ord = mIt->first.int2;
          int y23ord = mIt->first.int3;
          double presentcoeff = mIt->second;
          // ignore y12^(-n) silently except muq^(2+x)/y12
          // ignore y23^(-1) silently
          // require positive mass order
          if (
            y23ord >= 0 && muord >= 1 &&
            (y12ord >= 0 || (y12ord == -1 && muord >= 2) )
            ) {
            fin += 
              presentcoeff * pow(y12, y12ord)*pow(y23, y23ord) 
              * pow(mu1, muord);
          }
        }
        // add Mtilde terms
        for (
          map< FourInt,double>::iterator mtildeIt=LaurentMtilde.begin();
             mtildeIt!=LaurentMtilde.end();++mtildeIt
          ) {
          int muord = mtildeIt->first.int1;
          int y12ord = mtildeIt->first.int2;
          int y13ord = mtildeIt->first.int3;
          int y23ord = mtildeIt->first.int4;
          double presentcoeff = mtildeIt->second;
          // only mu^(2+x)/y12/y13 is allowed, other exponents of 
          // y12 and y13 are ignored silently
          if (
            y23ord >= 0 && y12ord == -1 && y13ord ==-1 && muord >= 2 
            ) {
            fin += 
              presentcoeff * pow(y12, y12ord) * pow(y13, y13ord) 
              * pow(y23, y23ord) * pow(mu1, muord);
          }
        }
      }
      
    } // end of if (useFiniteTerms)
    return eikonal + 1/den*(collinear + 1/den*fin);
  }
  
  
  
  // g-g emission antenna
  void AntennaggEmit::readC() {
    Antenna::readC();
    
    // read in collinear gluon splitting parameter
    // this is a global parameter, it cannot have different values for
    // different antennae
    const string cgsname("Vincia:CollinearGluonSplitting");
    if ( settingsPtr->isParm(cgsname) ) {	
      if ( setCollinearGluonSplitting(settingsPtr->parm(cgsname)) ) {
	int verb = (getCollinearGluonSplitting() == 0.0) ? 8 : 6; 
	if (verbose >= verb)
	  printOut(
            humanName+":readC",
            cgsname+" = "+num2str(getCollinearGluonSplitting(),6)
            );
      }
    }
  }
  
  void AntennaggEmit::clear(){
    if (inputMode == 1) {
      // Clear Laurent series coeffiecents
      Antenna::clear();
      setCollinearGluonSplitting(0.0);
    } else {
      // Enter alternative input modes here
    }
  }  
  
  void AntennaggEmit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "collinear gluon splitting: " << collinearGluonSplitting 
      << endl;
  }
  
  double AntennaggEmit::getCollinearGluonSplitting() {
    return collinearGluonSplitting;
  }
  
  bool AntennaggEmit::setCollinearGluonSplitting(double s) {
    if (inputMode == 1) {
      // check for positivity
      if (s >= 0.0) {
        collinearGluonSplitting = s;
        return true;
      }
      return false;
    }
    return false;
  }
  
  // rescaleFiniteTerms is inherited from Antenna
  
  // gettestMasses is inherited from Antenna
  
  double AntennaggEmit::a2Raw(
    double y12, double, double y23, double, double, double
    ) {
    double y13 = 1-y12-y23;
    double eikonal = 2*y13/(y12*y23);
    double collinear = 0.0;
    if (useCollinearTerms) {
      collinear = 
	y12/y23 + y23/y12 - pow2(y12)/y23 - pow2(y23)/y12
	+ collinearGluonSplitting*( (y12-y13)/y23 + (y23-y13)/y12);
    }
    double fin = 0.0;
    if (useFiniteTerms) {
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
        int ic=it->first;
        for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
          int jc=jt->first;  
          // restrict to combined exponent 2
          if (ic >= 0 && jc >= 0 && ic + jc <= 2) {
            double c=jt->second;
            // average with 1<->3 switched coefficient
            if (ic != jc) {
              c = 0.5*( c + getC(jc,ic) );
            }
            fin += c * pow(y12, ic) * pow(y23, jc); 
          }
        }      
      }
    }
    return eikonal + collinear + fin;
  }
  
  
  
  // q-g splitting antenna
  void AntennaqgSplit::readC() {
    Antenna::readC();
    // read also useMassTerms
    if ( settingsPtr->isFlag(name + ":UseMassTerms") ) {
      useMassTerms = settingsPtr->flag(name + ":UseMassTerms");
    }
    
    // use getParms
    // have to enter the bracket to not catch LaurentMtilde
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu1ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int mu23ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int y12ord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int y23ord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setM(mu1ord,mu23ord,y12ord,y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(mu1ord, mu23ord, y12ord, y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+" = "+num2str(getM(mu1ord, mu23ord, y12ord, y23ord),6)
          );
      }
    }
    // read Ctilde and Mtilde in analogy
    map<string,Parm> mtildecoeffs = 
      settingsPtr->getParmMap(name+":LaurentMtilde(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mtildecoeffs.begin();
      pt!=mtildecoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      size_t pos5 = parname.find(",", pos4+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu1ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int mu23ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int y12ord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      int y23ord = atoi(parname.substr(pos4+1, pos5-pos4-1).c_str());
      setMtilde(mu1ord,mu23ord,y12ord,y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getMtilde(mu1ord, mu23ord, y12ord, y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+
            " = "+num2str(getMtilde(mu1ord, mu23ord, y12ord, y23ord),6)
          );
      }
    }
    map<string,Parm> ctildecoeffs = 
      settingsPtr->getParmMap(name+":LaurentCtilde");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=ctildecoeffs.begin();
      pt!=ctildecoeffs.end();++pt
      ) {
      // read in the two numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int y12ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int y23ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      setCtilde(y12ord, y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getCtilde(y12ord, y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+
            " = "+num2str(getCtilde(y12ord, y23ord),6)
          );
      }
    }
  }
  
  void AntennaqgSplit::clear(){
    if (inputMode == 1) {
      // Clear Laurent series coeffiecents
      Antenna::clear();
      LaurentM.clear();
      LaurentMtilde.clear();
      LaurentCtilde.clear();
    } else {
      // Enter alternative input modes here
    }
  }  
  
  void AntennaqgSplit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    for (map<FourInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      FourInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mIt->second << endl;
    }
    for (
      map<FourInt, double>::iterator mtildeIt = LaurentMtilde.begin(); 
      mtildeIt != LaurentMtilde.end(); ++mtildeIt
      ) {
      FourInt indx = mtildeIt->first;
      os << "LaurentMtilde(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << ", " << indx.int4 << "): " 
        << mtildeIt->second << endl;
    }
    for (
      map<pair<int,int>, double>::iterator cIt = LaurentCtilde.begin(); 
      cIt != LaurentCtilde.end(); ++cIt
      ) {
      pair<int,int> indx = cIt->first;
      os << "LaurentCtilde(" << indx.first << ", " << indx.second 
        << "): " 
        << cIt->second << endl;
    }
  }
  
  bool AntennaqgSplit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    if (inputMode == 1) {
      // LaurentM
      for (map< FourInt, double>::iterator pt=LaurentM.begin();
	   pt!=LaurentM.end();++pt) {
        int muqord = pt->first.int1;
        int muqpord = pt->first.int2;
        int y12ord = pt->first.int3;
        int y23ord = pt->first.int4;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setM(
          muqord,muqpord, y12ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      // LaurentMtilde
      for (map< FourInt, double>::iterator pt=LaurentMtilde.begin();
	   pt!=LaurentMtilde.end();++pt) {
        int muqord = pt->first.int1;
        int muqpord = pt->first.int2;
        int y12ord = pt->first.int3;
        int y23ord = pt->first.int4;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setMtilde(
          muqord,muqpord, y12ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      // LaurentCtilde
      for (
        map< pair<int,int>, double>::iterator pt=LaurentCtilde.begin();
          pt!=LaurentCtilde.end();++pt
        ) {
        int y12ord = pt->first.first;
        int y23ord = pt->first.second;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setCtilde(y12ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      return baseOK;
    }
    else {
      if (verbose >= 3) {
        printOut(name+":rescaleFiniteTerms", 
                 "inputMode " + num2str( inputMode )
                 + " not recognized.");
      }
      return false;
    }
  }
  
  double AntennaqgSplit::getM(
    int muqord, int muqpord, int y12ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentM.find(FourInt(muqord, muqpord, y12ord, y23ord));
    if (pos == LaurentM.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaqgSplit::setM(
    int muqord, int muqpord, int y12ord, int y23ord, double mIn
    ) {
    if (inputMode == 1) {
      FourInt index(muqord, muqpord, y12ord, y23ord);
      if (mIn == 0.0) {
        LaurentM.erase(index);
      }
      else {
        LaurentM[index] = mIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setM",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  double AntennaqgSplit::getMtilde(
    int muqord, int muqpord, int y12ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map<FourInt, double>::iterator pos = 
      LaurentMtilde.find(FourInt(muqord, muqpord, y12ord, y23ord));
    if (pos == LaurentMtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaqgSplit::setMtilde(
    int muqord, int muqpord, int y12ord, int y23ord, double mIn
    ) {
    if (inputMode == 1) {
      FourInt index(muqord, muqpord, y12ord, y23ord);
      if (mIn == 0.0) {
        LaurentMtilde.erase(index);
      }
      else {
        LaurentMtilde[index] = mIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setM",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  double AntennaqgSplit::getCtilde(
    int y12ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map< pair<int,int>, double>::iterator pos = 
      LaurentCtilde.find(make_pair(y12ord, y23ord));
    if (pos == LaurentCtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaqgSplit::setCtilde(
    int y12ord, int y23ord, double cIn
    ) {
    if (inputMode == 1) {
      pair<int,int> index(y12ord, y23ord);
      if (cIn == 0.0) {
        LaurentCtilde.erase(index);
      }
      else {
        LaurentCtilde[index] = cIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setM",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  vector< vector<double> > AntennaqgSplit::gettestMasses() {
    const double e03_masses_input[][3] = 
    {	
      {0,0,0}, {0.01,0,0}, {0.05,0,0}, {0.3,0,0}, {0.5,0,0}, 
      {0.9,0,0}, {0.95,0,0}, {0.99,0,0}, {1,0,0},
      {0,0.01,0.01}, {0.01,0.01,0.01}, {0.05,0.01,0.01}, 
      {0.3,0.01,0.01}, {0.5,0.01,0.01}, {0.9,0.01,0.01}, 
      {0.95,0.01,0.01}, {0.98,0.01,0.01},
      {0,0.05,0.05}, {0.01,0.05,0.05}, {0.05,0.05,0.05}, 
      {0.3,0.05,0.05}, {0.5,0.05,0.05}, {0.85,0.05,0.05}, 
      {0.9,0.05,0.05}, 
      {0,0.3,0.3}, {0.01,0.3,0.3}, {0.05,0.3,0.3}, 
      {0.3,0.3,0.3}, {0.35,0.3,0.3}, {0.4,0.3,0.3}, 
      {0,0.45,0.45}, {0.01,0.45,0.45}, {0.05,0.45,0.45}, 
      {0.1,0.45,0.45}
    };
    vector< vector<double> > e03_masses;
    for (
      size_t i=0; i<sizeof(e03_masses_input)/(3*sizeof(double)); i++
      ) {
      vector<double> prov_masses;
      prov_masses.push_back(e03_masses_input[i][0]);
      prov_masses.push_back(e03_masses_input[i][1]);
      prov_masses.push_back(e03_masses_input[i][2]);
      e03_masses.push_back( prov_masses );
    }
    return e03_masses;
  }
  
  double AntennaqgSplit::a2Raw(
    double y12, double y13, double y23, 
    double muq, double muqp1, double muqp2
    ) {
    // call the function with mass zero if useMassTerms is false
    if ( !useMassTerms && muq != 0.0 && muqp1 != 0.0 && muqp2 != 0.0) {
      return a2Raw(y12, y13, y23, 0.0, 0.0, 0.0);
    }
    // silently average the secondary quark masses
    double muqp = 0.5*( muqp1 + muqp2 );
    double den = y23 + 2*pow2(muqp);
    // warning: if useMassTerms is false, a2Raw is called with a sum 
    // of reduced yab and muc which do NOT add up to 1
    double yIK = y12 + y13 + y23 + 2*pow2(muqp);
    double fixed = 
      0.5/den*( (pow2(y13) + pow2(y12)) / pow2(yIK) + 2*pow2(muqp)/den);
    double fin = 0.0;
    if (useFiniteTerms) {
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      // Add C(i,j) terms with prefactor y23/den
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
        int ic=it->first;
        for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
          int jc=jt->first;  
          // Only add finite terms
          if (ic >= 0 && jc >= 0) {
            double c=jt->second;      
            fin += y23/den * c * pow(y12,ic) * pow(y23,jc);
          }
        }      
      }
    }
    if (useFiniteTerms && (muq > 0.0 || muqp > 0.0) ) {
      // LaurentM
      for (
        map< FourInt,double>::iterator mIt=LaurentM.begin();
           mIt!=LaurentM.end();++mIt
        ) {
        int muqord = mIt->first.int1;
        int muqpord = mIt->first.int2;
        int y12ord = mIt->first.int3;
        int y23ord = mIt->first.int4;
        double presentcoeff = mIt->second;
        // only nonnegative powers of y12 and y23 are allowed
        // at least one positive power of muq or muqp is needed
        if (
          y12ord >= 0 && y23ord >= 0 && max(muqord, muqpord) >= 1 
          ) {
          fin += 
            y23/den * presentcoeff * pow(y12, y12ord) * pow(y23, y23ord) 
            * pow(muq, muqord) * pow(muqp, muqpord);
        }
      }
      // LaurentMtilde
      for (
        map< FourInt,double>::iterator mtildeIt=LaurentMtilde.begin();
           mtildeIt!=LaurentMtilde.end();++mtildeIt
        ) {
        int muqord = mtildeIt->first.int1;
        int muqpord = mtildeIt->first.int2;
        int y12ord = mtildeIt->first.int3;
        int y23ord = mtildeIt->first.int4;
        double presentcoeff = mtildeIt->second;
        // only nonnegative powers of y12 and y23 are allowed
        // at least one positive power of muq or muqp is needed
        if (
          y12ord >= 0 && y23ord >= 0 && max(muqord, muqpord) >= 1 
          ) {
          fin += 
            pow2(muqp)/den * presentcoeff * pow(y12, y12ord) * 
            pow(y23, y23ord) * pow(muq, muqord) * pow(muqp, muqpord);
        }
      }
      // LaurentCtilde
      for (
        map< pair<int,int>, double>::iterator cIt=LaurentCtilde.begin(); 
        cIt != LaurentCtilde.end(); ++cIt
        ) {
        int y12ord = cIt->first.first;
        int y23ord = cIt->first.second;
        double presentcoeff = cIt->second;
        // only nonnegative powers are allowed
        if (y12ord >= 0 && y23ord >= 0) {
          fin += 
            pow2(muqp)/den * presentcoeff * pow(y12, y12ord) * 
            pow(y23, y23ord);
        }
      }
    }
    return fixed + fin/pow2(yIK);
  }
  
  
  
  // g-g splitting antenna
  void AntennaggSplit::readC() {
    Antenna::readC();
    // read also useMassTerms
    if ( settingsPtr->isFlag(name + ":UseMassTerms") ) {
      useMassTerms = settingsPtr->flag(name + ":UseMassTerms");
    }
    // read Ctilde and M
    map<string,Parm> mcoeffs = 
      settingsPtr->getParmMap(name+":LaurentM(");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=mcoeffs.begin();
      pt!=mcoeffs.end();++pt
      ) {
      // read in the four numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      size_t pos4 = parname.find(",", pos3+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int mu23ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int y12ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      int y23ord = atoi(parname.substr(pos3+1, pos4-pos3-1).c_str());
      setM(mu23ord,y12ord,y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getM(mu23ord, y12ord, y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+
            " = "+num2str(getM(mu23ord, y12ord, y23ord),6)
          );
      }
    }
    map<string,Parm> ctildecoeffs = 
      settingsPtr->getParmMap(name+":LaurentCtilde");
    // loop through the map, add the entries
    for (
      map<string,Parm>::iterator pt=ctildecoeffs.begin();
      pt!=ctildecoeffs.end();++pt
      ) {
      // read in the two numbers from the string
      string parname = pt->first;
      // determine the positions of the numbers
      size_t pos1 = parname.find("(");
      size_t pos2 = parname.find(",", pos1+1);
      size_t pos3 = parname.find(",", pos2+1);
      // conversion should be done in a safer way, but for the moment
      // this should do it
      int y12ord = atoi(parname.substr(pos1+1, pos2-pos1-1).c_str());
      int y23ord = atoi(parname.substr(pos2+1, pos3-pos2-1).c_str());
      setCtilde(y12ord, y23ord, pt->second.valNow);
      // output
      int verb = 
        (abs(getCtilde(y12ord, y23ord)) <= TINY) ? 8 : 6; 
      if (verbose >= verb) {
	printOut(
          humanName+":readC",
          parname+
            " = "+num2str(getCtilde(y12ord, y23ord),6)
          );
      }
    }
  }
  
  void AntennaggSplit::clear(){
    if (inputMode == 1) {
      // Clear Laurent series coeffiecents
      Antenna::clear();
      LaurentM.clear();
      LaurentCtilde.clear();
    } else {
      // Enter alternative input modes here
    }
  }  
  
  void AntennaggSplit::listCoefficients(ostream& os) {
    Antenna::listCoefficients(os);
    os << "use mass terms: "; 
    useMassTerms ? os << "True" : os << "False"; 
    os << endl;
    for (
      map<ThreeInt, double>::iterator mIt = LaurentM.begin(); 
      mIt != LaurentM.end(); ++mIt
      ) {
      ThreeInt indx = mIt->first;
      os << "LaurentM(" << indx.int1 << ", " << indx.int2 << ", " 
        << indx.int3 << "): " 
        << mIt->second << endl;
    }
    for (
      map<pair<int,int>, double>::iterator cIt = LaurentCtilde.begin(); 
      cIt != LaurentCtilde.end(); ++cIt
      ) {
      pair<int,int> indx = cIt->first;
      os << "LaurentCtilde(" << indx.first << ", " << indx.second 
        << "): " 
        << cIt->second << endl;
    }
  }
  
  bool AntennaggSplit::rescaleFiniteTerms(double rescalingFactor) {
    bool baseOK = Antenna::rescaleFiniteTerms(rescalingFactor);
    if (inputMode == 1) {
      // LaurentM
      for (map< ThreeInt, double>::iterator pt=LaurentM.begin();
	   pt!=LaurentM.end();++pt) {
        int muqpord = pt->first.int1;
        int y12ord = pt->first.int2;
        int y23ord = pt->first.int3;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setM(
          muqpord, y12ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      // LaurentCtilde
      for (
        map< pair<int,int>, double>::iterator pt=LaurentCtilde.begin();
          pt!=LaurentCtilde.end();++pt
        ) {
        int y12ord = pt->first.first;
        int y23ord = pt->first.second;
        double oldcoeff = pt->second;
        double newabs = rescalingFactor*abs(oldcoeff);
        // round the new absolute value down to the closest
        // multiple of abs(1-rescalingFactor)
        double minchange = abs(1-rescalingFactor);
        newabs = floor( newabs/minchange )*minchange;
        setCtilde(y12ord, y23ord, oldcoeff*newabs/abs(oldcoeff));
      }
      return baseOK;
    }
    else {
      if (verbose >= 3) {
        printOut(name+":rescaleFiniteTerms", 
                 "inputMode " + num2str( inputMode )
                 + " not recognized.");
      }
      return false;
    }
  }
  
  double AntennaggSplit::getM(
    int muqpord, int y12ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map<ThreeInt, double>::iterator pos = 
      LaurentM.find(ThreeInt(muqpord, y12ord, y23ord));
    if (pos == LaurentM.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaggSplit::setM(
    int muqpord, int y12ord, int y23ord, double mIn
    ) {
    if (inputMode == 1) {
      ThreeInt index(muqpord, y12ord, y23ord);
      if (mIn == 0.0) {
        LaurentM.erase(index);
      }
      else {
        LaurentM[index] = mIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setM",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  double AntennaggSplit::getCtilde(
    int y12ord, int y23ord
    ) {
    // use find to avoid the creation of an entry
    map< pair<int,int>, double>::iterator pos = 
      LaurentCtilde.find(make_pair(y12ord, y23ord));
    if (pos == LaurentCtilde.end()) {
      return 0.0;
    }
    else {
      return pos->second;
    }
  }
  
  bool AntennaggSplit::setCtilde(
    int y12ord, int y23ord, double cIn
    ) {
    if (inputMode == 1) {
      pair<int,int> index(y12ord, y23ord);
      if (cIn == 0.0) {
        LaurentCtilde.erase(index);
      }
      else {
        LaurentCtilde[index] = cIn;
      }
    }
    else {
      // Enter alternative input modes here
      if (verbose >= 1)
        printErr(humanName+":setM",
               "sorry, desired inputMode not recognized.");
      return false;
    }
    return true;
  }
  
  vector< vector<double> > AntennaggSplit::gettestMasses() {
    const double g03_masses_input[] =
      { 0., 0.01, 0.05, 0.1, 0.3, 0.45, 0.49, 0.5 };
    vector< vector<double> > g03_masses;
    for (size_t i=0; i<sizeof(g03_masses_input)/sizeof(double); i++) {
      vector<double> prov_masses;
      prov_masses.push_back(0);
      prov_masses.push_back(g03_masses_input[i]);
      prov_masses.push_back(g03_masses_input[i]);
      g03_masses.push_back( prov_masses );
    }
    return g03_masses;
  }
  
  double AntennaggSplit::a2Raw(
    double y12, double y13, double y23, 
    double, double muqp1, double muqp2
    ) {
    // call the function with mass zero if useMassTerms is false
    if ( !useMassTerms && muqp1 != 0.0 && muqp2 != 0.0) {
      return a2Raw(y12, y13, y23, 0.0, 0.0, 0.0);
    }
    // silently average the secondary quark masses
    double muqp = 0.5*( muqp1 + muqp2 );
    double den = y23 + 2*pow2(muqp);
    double fixed = 
      0.5/den*( pow2(y13) + pow2(y12) + 2*pow2(muqp)/den);
    double fin = 0.0;
    if (useFiniteTerms) {
      map<int,map<int,double> >::iterator it;
      map<int,double>::iterator jt;
      // Add C(i,j) terms with prefactor y23/den
      for (it=LaurentC.begin();it!=LaurentC.end();++it) {
        int ic=it->first;
        for (jt=LaurentC[ic].begin();jt!=LaurentC[ic].end();++jt) {
          int jc=jt->first;  
          // Only add finite terms
          if (ic >= 0 && jc >= 0) {
            double c=jt->second;      
            fin += y23/den * c * pow(y12,ic) * pow(y23,jc);
          }
        }      
      }
    }
    if ( useFiniteTerms && muqp > 0.0 ) {
      // LaurentM
      for (
        map< ThreeInt,double>::iterator mIt=LaurentM.begin();
           mIt!=LaurentM.end();++mIt
        ) {
        int muqpord = mIt->first.int1;
        int y12ord = mIt->first.int2;
        int y23ord = mIt->first.int3;
        double presentcoeff = mIt->second;
        // only nonnegative powers of y12 and y23 are allowed
        // at least one positive power of muq is needed
        if (
          y12ord >= 0 && y23ord >= 0 && muqpord >= 1 
          ) {
          fin += 
            pow2(muqp)/den * presentcoeff * pow(y12, y12ord) 
            * pow(y23, y23ord) * pow(muqp, muqpord);
        }
      }
      // LaurentCtilde
      for (
        map< pair<int,int>, double>::iterator cIt=LaurentCtilde.begin();
        cIt != LaurentCtilde.end(); ++cIt
        ) {
        int y12ord = cIt->first.first;
        int y23ord = cIt->first.second;
        double presentcoeff = cIt->second;
        // only nonnegative powers are allowed
        if (y12ord >= 0 && y23ord >= 0) {
          fin += 
            pow2(muqp)/den * presentcoeff * pow(y12, y12ord) * 
            pow(y23, y23ord);
        }
      }
    }
    return fixed + fin;
  }

  
} // End Vincia Namespace
  
